"""Quantum arithmetic operations for quantum states."""

# SPDX-FileCopyrightText: 2025 Evandro Chagas Ribeiro da Rosa <evandro@quantuloop.com>
#
# SPDX-License-Identifier: Apache-2.0

from math import pi
from ..base import Quant
from ..gates import QFT, P, X
from ..operations import ctrl, around

__all__ = [
    "addi",
    "mul",
    "set_int",
]


def _rotation(l, q, m=1):
    P(2 * pi / (2**l / m), q)


def _addi_qq(lhs: Quant, rhs: Quant, m=1):
    n_lhs = len(lhs)
    n_rhs = len(rhs)
    if n_rhs > n_lhs:
        raise RuntimeError(
            "The number of qubits in the right-hand side must be"
            "less than or equal to the left-hand side."
        )
    if n_lhs == 0:
        return

    offset = n_lhs - n_rhs + 1

    for j, q in enumerate(rhs):
        ctrl(q, _rotation)(j + offset, lhs[0], m)

    if n_rhs == n_lhs:
        _addi_qq(lhs[1:], rhs[1:], m)
    else:
        _addi_qq(lhs[1:], rhs, m)


def _addi_qi(lhs: Quant, rhs: int, m=1):
    n_lhs = len(lhs)
    n_rhs = rhs.bit_length()
    if n_rhs > n_lhs:
        raise RuntimeError(
            "The number of bits in the right-hand side must be"
            "less than or equal to the left-hand side."
        )
    if n_lhs == 0:
        return

    offset = n_lhs - n_rhs + 1

    for j, q in enumerate(f"{rhs:0{n_rhs}b}"):
        if q == "1":
            _rotation(j + offset, lhs[0], m)

    if n_rhs == n_lhs:
        _addi_qi(lhs[1:], rhs & ((1 << n_rhs - 1) - 1), m)
    else:
        _addi_qi(lhs[1:], rhs, m)


def addi(lhs: Quant, rhs: Quant | int, multiplier: int = 1):
    """Add two quantum states or a quantum state and an integer.

    The result is stored in the left-hand side quantum state.

    The multiplier is used to scale the right-hand side.

    Args:
        lhs: Left-hand side quantum state.
        rhs: Right-hand side quantum state or integer to add.
        multiplier: Multiplier for the addition (default is 1).
    """

    with around(QFT, reversed(lhs), False):
        if isinstance(rhs, Quant):
            _addi_qq(lhs, rhs, multiplier)
        elif isinstance(rhs, int):
            _addi_qi(lhs, rhs, multiplier)
        else:
            raise TypeError("Right-hand side must be a Quant or an integer.")


def mul(lhs: Quant, rhs: Quant | int, result: Quant):
    """Multiply two quantum states.

    The result is stored in the result quantum state.

    Args:
        lhs: Left-hand side quantum state.
        rhs: Right-hand side quantum state or integer to add.
        result: Quantum state to store the result.
    """

    for i, q in enumerate(reversed(lhs)):
        ctrl(q, addi)(result, rhs, 2**i)


def set_int(qubits: Quant, state: int):
    r"""Set the state of a quantum register.

    If the qubits are in the state :math:`\ket{0\dots0}`, this function will
    set them to the state :math:`\ket{\texttt{state}}`.

    Args:
        state: Integer representing the state to set.
        qubits: Quantum register to modify.
    """

    for i, qubit in zip(f"{state:0{len(qubits)}b}", qubits):
        if i == "1":
            X(qubit)



"""Expected value calculation utilities.

This module provides base classes creating a Hamiltonian fro expected value calculation.
"""

from __future__ import annotations

# SPDX-FileCopyrightText: 2020 Evandro Chagas Ribeiro da Rosa <evandro@quantuloop.com>
# SPDX-FileCopyrightText: 2020 Rafael de Santiago <r.santiago@ufsc.br>
#
# SPDX-License-Identifier: Apache-2.0

# pylint: disable=duplicate-code

from ctypes import c_int32, c_size_t
from numbers import Number
from itertools import product
from functools import reduce
from operator import add
from typing import Literal

from .base import Process, Quant

from .clib.libket import API

__all__ = [
    "Pauli",
    "Hamiltonian",
    "commutator",
    "ExpValue",
]


class Pauli:
    """Pauli operator for Hamiltonian creation.

    .. tip::
        The preferred way to create a Hamiltonian is by using the
        :func:`~ket.gates.ham` context manager, which allows you to define a Hamiltonian
        using a more intuitive syntax. However, you can also create Hamiltonians directly by
        instantiating this class.

    This class represents a Pauli operator for Hamiltonian creation. The primary usage of this class
    is to prepare a Hamiltonian by adding and multiplying Pauli operators and scalars for
    calculating the expected value of a quantum state.

    Allowed operations:
    - Multiply by a scalar or another :class:`~ket.expv.Pauli` operator.
    - Add another :class:`~ket.expv.Pauli` or :class:`~ket.expv.Hamiltonian` operator.

    Example:

        .. code-block::

            from ket import *

            p = Process()
            qubits = p.alloc(3)
            H(qubits[1])
            S(H(qubits[2]))

            # Example 1: Single Pauli term
            pauli_term = Pauli("X", qubits[0])
            print(repr(pauli_term))
            # <Ket 'Pauli' 1.0*X0, pid=0x...>

            # Example 2: Sum of Pauli terms
            sum_pauli = Pauli("Y", qubits[1]) + Pauli("Z", qubits[2])
            print(repr(sum_pauli))
            # <Ket 'Hamiltonian' 1.0*Y1 + 1.0*Z2, pid=0x...>

            # Example 3: Pauli multiplication
            multiplied_pauli = 2.0 * Pauli("X", qubits[0]) * Pauli("Y", qubits[1])
            print(repr(multiplied_pauli))
            # <Ket 'Pauli' 2.0*X0Y1, pid=0x...>

            # Example 4: Calculating expected value
            h = Pauli("Z", qubits[0]) + Pauli("X", qubits[1]) + Pauli("Y", qubits[2])
            print(repr(h))
            # <Ket 'Hamiltonian' 1.0*Z0 + 1.0*X1 + 1.0*Y2, pid=0x...>
            result = exp_value(h)
            print(result.value)  # 3.0000000000000013


    Args:
        pauli: Pauli operator type.
        qubits: Qubits to apply the Pauli operator to.
        process: *For internal usage*. Quantum process, default is the process of the given qubits.
        pauli_list: *For internal usage*. List of Pauli operators.
        qubits_list: *For internal usage*. List of Qubit.
        coef: *For internal usage*. Coefficient for the Pauli operator, default is 1.0.

    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        pauli: Literal["X", "Y", "Z", "I"],
        qubits: Quant,
        *,
        _process: Process | None = None,
        _map: dict[int, str] | None = None,
        _coef: float | None = None,
    ):
        if qubits is not None and not isinstance(qubits, Quant):
            qubits = reduce(add, qubits)

        if pauli is not None and qubits is not None:
            self.process = qubits.process
            self.map = {q: pauli for q in qubits.qubits}
            self.coef = 1.0
        else:
            assert _process is not None and _map is not None and _coef is not None, (
                "If pauli and qubits are not provided, "
                "process, map, and coef must be provided."
            )

            self.process = _process
            self.map = _map
            self.coef = _coef

    def __neg__(self) -> Pauli:
        return -1.0 * self

    def __mul__(self, other: float | Pauli) -> Pauli:

        if isinstance(other, Number):
            return Pauli(
                None,
                None,
                _process=self.process,
                _map=self.map,
                _coef=self.coef * other,
            )

        if isinstance(other, Hamiltonian):
            return other.__mul__(self)

        if self.process is not other.process:
            raise ValueError("different Ket processes")

        if not set(self.map.keys()).isdisjoint(other.map.keys()):
            raise ValueError("Pauli operators act on the same qubits")

        return Pauli(
            None,
            None,
            _process=self.process,
            _map={**self.map, **other.map},
            _coef=self.coef * other.coef,
        )

    def __matmul__(self, rhs: Pauli) -> Pauli:
        if isinstance(rhs, Hamiltonian):
            return Hamiltonian([self], self.process) @ rhs

        if self.process is not rhs.process:
            raise ValueError("different Ket processes")

        qubits = {*self.map.keys(), *rhs.map.keys()}
        result_map = {}
        coef = self.coef * rhs.coef

        for qubit in qubits:
            match (self.map.get(qubit, "I"), rhs.map.get(qubit, "I")):
                case ("I", "I"):
                    pass
                case ("I", p) | (p, "I"):
                    result_map[qubit] = p
                case ("X", "Y"):
                    result_map[qubit] = "Z"
                    coef *= 1j
                case ("X", "X"):
                    result_map[qubit] = "I"
                case ("X", "Z"):
                    result_map[qubit] = "Y"
                    coef *= -1j
                case ("Y", "X"):
                    result_map[qubit] = "Z"
                    coef *= -1j
                case ("Y", "Y"):
                    result_map[qubit] = "I"
                case ("Y", "Z"):
                    result_map[qubit] = "X"
                    coef *= 1j
                case ("Z", "X"):
                    result_map[qubit] = "Y"
                    coef *= 1j
                case ("Z", "Y"):
                    result_map[qubit] = "X"
                    coef *= -1j
                case ("Z", "Z"):
                    result_map[qubit] = "I"
                case _:
                    raise ValueError("Unsupported Pauli multiplication")

        return Pauli(
            None,
            None,
            _process=self.process,
            _map=result_map,
            _coef=coef,
        )

    __rmul__ = __mul__

    def __len__(self) -> int:
        return len(self.map)

    def __add__(self, other) -> Hamiltonian:
        if isinstance(other, Number):
            other = Pauli(None, None, _process=self.process, _map={}, _coef=other)

        if self.process is not other.process:
            raise ValueError("different Ket processes")

        if isinstance(other, Hamiltonian):
            result = Hamiltonian([self], process=self.process) + other
        else:
            result = Hamiltonian([self, other], process=self.process)

        result._filter()
        return result

    __radd__ = __add__

    def __sub__(self, other) -> Hamiltonian:
        return self + -other

    def __rsub__(self, other) -> Hamiltonian:
        return -self + other

    def __truediv__(self, other: float) -> Pauli:
        return self.__mul__(1.0 / other)

    @staticmethod
    def x(qubits: Quant) -> Pauli:
        """Pauli X operator.

        Args:
            qubits: Qubits to apply the Pauli operator to.
        """
        return Pauli("X", qubits)

    @staticmethod
    def y(qubits: Quant) -> Pauli:
        """Pauli Y operator.

        Args:
            qubits: Qubits to apply the Pauli operator to.
        """
        return Pauli("Y", qubits)

    @staticmethod
    def z(qubits: Quant) -> Pauli:
        """Pauli Z operator.

        Args:
            qubits: Qubits to apply the Pauli operator to.
        """
        return Pauli("Z", qubits)

    @staticmethod
    def i(qubits: Quant) -> Pauli:
        """Pauli I operator.

        Args:
            qubits: Qubits to apply the Pauli I operator to.
        """
        return Pauli("I", qubits)

    def _str_no_coef(self) -> str:
        return "".join(f"{pauli}{qubit}" for qubit, pauli in sorted(self.map.items()))

    def __str__(self) -> str:
        return f"{self.coef}" + ("*" + self._str_no_coef() if len(self.map) > 0 else "")

    def __repr__(self) -> str:
        return f"<Ket 'Pauli' {str(self)}, pid={hex(id(self.process))}>"


class Hamiltonian:
    """Hamiltonian for expected value calculation.

    This class is not intended to be instantiated directly. Instead, it should be created
    by adding :class:`~ket.expv.Pauli` operators or other :class:`~ket.expv.Hamiltonian`
    objects.
    """

    def __init__(self, pauli_products: list[Pauli], process: Process):
        self.process = process
        self.pauli_products = pauli_products
        assert all(isinstance(p, Pauli) for p in pauli_products)

    def __add__(self, other: Hamiltonian | Pauli) -> Hamiltonian:
        if isinstance(other, Number):
            other = Pauli(None, None, _process=self.process, _map={}, _coef=other)
        if isinstance(other, Pauli):
            other = Hamiltonian([other], self.process)

        if self.process is not other.process:
            raise ValueError("different Ket processes")

        result = Hamiltonian(self.pauli_products + other.pauli_products, self.process)
        result._filter()

        return result

    def __sub__(self, other: Hamiltonian | Pauli) -> Hamiltonian:
        return self + -other

    __radd__ = __add__

    def _filter(self):
        new_terms = {}
        for term in self.pauli_products:
            str_term = term._str_no_coef()  # pylint: disable=protected-access
            if str_term not in new_terms:
                new_terms[str_term] = term
            else:
                new_terms[str_term].coef += term.coef

        self.pauli_products = [
            term for term in new_terms.values() if abs(term.coef) > 1e-10
        ]

    def __matmul__(self, other: Hamiltonian | Pauli) -> Hamiltonian:
        if isinstance(other, Pauli):
            other = Hamiltonian([other], self.process)

        result = Hamiltonian(
            [a @ b for a, b in product(self.pauli_products, other.pauli_products)],
            self.process,
        )

        result._filter()

        return result

    def __rsub__(self, other) -> Hamiltonian:
        return -self + other

    def __mul__(self, other: float | Hamiltonian | Pauli) -> Hamiltonian:
        if isinstance(other, (Number, Pauli)):
            return Hamiltonian(
                [p * other for p in self.pauli_products], process=self.process
            )

        if isinstance(other, Hamiltonian):
            return Hamiltonian(
                [p * q for p, q in product(self.pauli_products, other.pauli_products)],
                process=self.process,
            )

        raise TypeError(
            f"Unsupported type for multiplication: {type(other)}. "
            "Expected float, Hamiltonian, or Pauli."
        )

    __rmul__ = __mul__

    def __truediv__(self, other: float) -> Hamiltonian:
        return self.__mul__(1.0 / other)

    def __neg__(self) -> Hamiltonian:
        return -1.0 * self

    def __len__(self) -> int:
        return len(self.pauli_products)

    def __repr__(self) -> str:
        return (
            f"<Ket 'Hamiltonian' {' + '.join(str(p) for p in self.pauli_products)}, "
            f"pid={hex(id(self.process))}>"
        )


def commutator(a: Hamiltonian, b: Hamiltonian) -> Hamiltonian:
    """Calculate the commutator of two Hamiltonians."""
    return a @ b - b @ a


class ExpValue:
    """Expected value for a quantum state.

    This class holds a reference for a expected value result. The result may not be available right
    after the measurement call, especially in batch execution.

    To read the value, access the attribute :attr:`~ket.base.ExpValue.value`. If the value is not
    available, the measurement will return `None`; otherwise, it will return a float.

    You can instantiate this class by calling the :func:`~ket.operations.exp_value` function.

    Example:

        .. code-block:: python

            from ket import *

            p = Process()
            q = p.alloc(2)
            CNOT(H(q[0]), q[1])
            result = exp_value(Pauli("X", q))
            print(result.value) # 1.0000000000000004

    """

    pauli_map = {"X": 1, "Y": 2, "Z": 3}

    def __init__(self, hamiltonian: Hamiltonian | Pauli):
        if isinstance(hamiltonian, Pauli):
            hamiltonian = Hamiltonian([hamiltonian], process=hamiltonian.process)

        self.process = hamiltonian.process

        hamiltonian_ptr = API["ket_hamiltonian_new"]()
        products_count = 0
        self._i_coef = 0.0
        for pauli_product in hamiltonian.pauli_products:
            if len(pauli_product.map) == 0:
                self._i_coef += pauli_product.coef.real
                continue

            qubits, pauli = zip(*pauli_product.map.items())
            pauli_qubits = list(
                zip(
                    *[(self.pauli_map[p], q) for p, q in zip(pauli, qubits) if p != "I"]
                )
            )
            if (
                isinstance(pauli_product.coef, complex)
                and abs(pauli_product.coef.imag) > 1e-10
            ):
                raise ValueError(
                    "Complex coefficients are not supported in Hamiltonian"
                    " expected value calculation."
                )

            if not pauli_qubits:
                self._i_coef += pauli_product.coef.real
                continue
            pauli, qubits = pauli_qubits
            API["ket_hamiltonian_add"](
                hamiltonian_ptr,
                (c_int32 * len(pauli))(*pauli),
                len(pauli),
                (c_size_t * len(qubits))(*qubits),
                len(qubits),
                pauli_product.coef.real,
            )
            products_count += 1

        if products_count > 0:
            self.index = self.process.exp_value(hamiltonian_ptr).value
            self._value = None
        else:
            self._value = self._i_coef

    def _check(self):
        if self._value is None:
            available, value = self.process.get_exp_value(self.index)
            if available.value:
                self._value = value.value + self._i_coef

    @property
    def value(self) -> float | None:
        """Retrieve the expected value if available."""
        self._check()
        return self._value

    def get(self) -> float:
        """Retrieve the expected values.

        If the value is not available, the quantum process will execute to get the result.
        """

        self._check()
        if self._value is None:
            self.process.execute()
        return self.value

    def __repr__(self) -> str:
        return f"<Ket 'ExpValue' value={self.value}, pid={hex(id(self.process))}>"



"""Quantum state snapshot.

This module provides the base class for a snapshot of a quantum state obtained
using a quantum simulator.
"""

from __future__ import annotations

# SPDX-FileCopyrightText: 2024 Evandro Chagas Ribeiro da Rosa <evandro@quantuloop.com>
#
# SPDX-License-Identifier: Apache-2.0

from math import pi
from random import Random
from cmath import sqrt, phase
from collections import defaultdict
from typing import Literal
from ctypes import c_size_t

from .base import Quant, _check_visualize

try:
    import numpy as np
    import plotly.graph_objs as go
    import plotly.express as px
except ImportError:
    pass

try:
    from IPython import get_ipython
    from IPython.display import Math

    try:
        import google.colab  # pylint: disable=unused-import

        IN_NOTEBOOK = True
    except ImportError:
        IN_NOTEBOOK = get_ipython().__class__.__name__ == "ZMQInteractiveShell"
except ImportError:
    IN_NOTEBOOK = False

__all__ = ["QuantumState"]


class QuantumState:
    """Snapshot of a quantum state.

    This class holds a reference for a snapshot of a quantum state obtained using a quantum
    simulator. The result may not be available right after the measurement call, especially in batch
    execution.

    You can instantiate this class by calling the :func:`~ket.operations.dump` function.

    :Example:

        .. code-block:: python

            from ket import *

            p = Process()
            a, b = p.alloc(2)
            with around(cat(kron(H, I), CNOT), a, b):
                Y(a)
                inside = dump(a+b)

            outside = dump(a+b)
            print(inside.show())
            # |01⟩    (50.00%)
            #  -0.707107i     ≅     -i/√2
            # |10⟩    (50.00%)
            #  0.707107i     ≅      i/√2

            print(outside.show())
            # |11⟩    (100.00%)
            #  -1.000000i     ≅     -i/√1

    Args:
        qubits: Qubits from which to capture a quantum state snapshot.
    """

    def __init__(self, qubits: Quant):
        self.qubits = qubits
        self.process = qubits.process
        self.index = self.process.dump(
            (c_size_t * len(qubits.qubits))(*qubits.qubits), len(qubits.qubits)
        ).value
        self.size = len(qubits)
        self._states = None

    def _get_ket_process(self):
        return self.process

    def _check(self):
        if self._states is None:
            available, size = self.process.get_dump_size(self.index)
            if available.value:
                states = defaultdict(complex)
                for i in range(size.value):
                    state, state_size, amp_real, amp_imag = self.process.get_dump(
                        self.index, i
                    )
                    state = int(
                        "".join(f"{state[j]:064b}" for j in range(state_size.value)), 2
                    )
                    amplitude = complex(amp_real.value, amp_imag.value)
                    if abs(amplitude) ** 2 > 1e-10:
                        states[state] += amplitude

                p = sum(map(lambda a: abs(a) ** 2, states.values()))
                if abs(p - 1.0) < 1e-10:
                    self._states = dict(states)
                else:
                    self._states = {
                        state: amplitude / sqrt(p) if p != 0 else 0
                        for state, amplitude in states.items()
                        if abs(amplitude) ** 2 > 1e-10
                    }

    @property
    def states(self) -> dict[int, complex] | None:
        """Get the quantum state.

        Returns a dictionary mapping base states to their corresponding probability amplitudes.

        Returns:
            The quantum state, or None if the quantum state information is not available.
        """

        self._check()
        return self._states

    def get(self) -> dict[int, complex]:
        """Get the quantum state.

        If the quantum state is not available, the quantum process will execute to get the result.
        """

        self._check()
        if self._states is None:
            self.process.execute()
        return self.states

    @property
    def probabilities(self) -> dict[int, float] | None:
        """Get the measurement probabilities.

        Returns a dictionary mapping base states to their corresponding measurement probabilities.

        Returns:
            The measurement probabilities, or None if the quantum state information is not
            available.
        """
        self._check()
        if self._states is None:
            return None
        return dict(map(lambda a: (a[0], abs(a[1]) ** 2), self._states.items()))

    def sample(self, shots=4096, seed=None) -> dict[int, int] | None:
        """Get the quantum execution shots.

        The parameters ``shots`` and ``seed`` are used to generate the sample from the quantum state
        snapshot.

        Args:
            shots: Number of shots. Defaults to 4096.
            seed: Seed for the RNG.

        Returns:
            A dictionary mapping measurement outcomes to their frequencies in the generated sample,
            or None if the quantum state information is not available.
        """

        self._check()
        if self._states is None:
            return None

        rng = Random(seed)
        shots = rng.choices(list(self.states), list(self.probabilities), k=shots)
        result = {}
        for state in shots:
            if state not in result:
                result[state] = 1
            else:
                result[state] += 1
        return result

    @staticmethod
    def _sphere():  # pylint: disable=too-many-locals
        phi = np.linspace(0, np.pi, 20)
        theta = np.linspace(0, 2 * np.pi, 40)
        phi, theta = np.meshgrid(phi, theta)
        x = np.sin(phi) * np.cos(theta)
        y = np.sin(phi) * np.sin(theta)
        z = np.cos(phi)
        sphere = go.Surface(
            x=x, y=y, z=z, showscale=False, opacity=0.02, name="Bloch Sphere"
        )

        equator_theta = np.linspace(0, 2 * np.pi, 100)
        equator_x = np.cos(equator_theta)
        equator_y = np.sin(equator_theta)
        equator_z = np.zeros_like(equator_theta)

        equator = go.Scatter3d(
            x=equator_x,
            y=equator_y,
            z=equator_z,
            mode="lines",
            line={"color": "gray", "width": 3},
            opacity=0.1,
            name="Equator",
        )

        z_line = go.Scatter3d(
            x=[0, 0],
            y=[0, 0],
            z=[1, -1],
            mode="lines",
            line={"color": "gray", "width": 3},
            opacity=0.1,
            name="z line",
        )

        x_line = go.Scatter3d(
            x=[1, -1],
            y=[0, 0],
            z=[0, 0],
            mode="lines",
            line={"color": "gray", "width": 3},
            opacity=0.1,
            name="x line",
        )

        y_line = go.Scatter3d(
            x=[0, 0],
            y=[1, -1],
            z=[0, 0],
            mode="lines",
            line={"color": "gray", "width": 3},
            opacity=0.1,
            name="y line",
        )

        basis_points = [
            ([0.0, 0.0, 1.0], "|0⟩"),
            ([0.0, 0.0, 0.8], "Z"),
            ([0.0, 0.0, -1.0], "|1⟩"),
            ([1.0, 0.0, 0.0], "|+⟩"),
            ([0.8, 0.0, 0.0], "X"),
            ([-1.0, 0.0, 0.0], "|‒⟩"),
            ([0.0, 1.0, 0.0], "|+i⟩"),
            ([0.0, 0.8, 0.0], "Y"),
            ([0.0, -1.0, 0.0], "|-i⟩"),
        ]

        basis = [
            go.Scatter3d(
                x=[p[0]],
                y=[p[1]],
                z=[p[2]],
                mode="text",
                text=[text],
                textposition="middle center",
                name=text,
            )
            for p, text in basis_points
        ]

        return [
            sphere,
            equator,
            x_line,
            y_line,
            z_line,
            *basis,
        ]

    def sphere(self) -> go.Figure:
        """Generate a Bloch sphere plot representing the quantum state.

        This method creates a Bloch sphere plot visualizing of one qubit quantum state.


        Note:
            This method requires additional dependencies from ``ket-lang[plot]``.

            Install with: ``pip install ket-lang[plot]``.

        Returns:
            A Bloch sphere plot illustrating the quantum state.
        """

        if len(self.qubits) != 1:
            raise ValueError("Bloch sphere plot is available only for 1 qubit")
        _check_visualize()

        ket = np.array(
            [
                [self.get()[0] if 0 in self.get() else 0.0],
                [self.get()[1] if 1 in self.get() else 0.0],
            ]
        )

        bra = np.conjugate(ket.T)

        pauli_x = np.array([[0, 1], [1, 0]])
        pauli_y = np.array([[0, -1j], [1j, 0]])
        pauli_z = np.array([[1, 0], [0, -1]])
        exp_x = (bra @ pauli_x @ ket).item().real
        exp_y = (bra @ pauli_y @ ket).item().real
        exp_z = (bra @ pauli_z @ ket).item().real

        qubit = go.Scatter3d(
            x=[exp_x],
            y=[exp_y],
            z=[exp_z],
            mode="markers",
            marker={"size": 5, "color": "red"},
            name="qubit",
        )

        line = go.Scatter3d(
            x=[0, exp_x],
            y=[0, exp_y],
            z=[0, exp_z],
            mode="lines",
            line={"color": "red", "width": 3},
            opacity=0.5,
            name="qubit line",
        )

        fig = go.Figure(
            data=[
                *self._sphere(),
                qubit,
                line,
            ]
        )

        fig.update_layout(
            scene={
                "xaxis": {
                    "range": [-1, 1],
                    "showgrid": False,
                    "showbackground": False,
                    "visible": False,
                },
                "yaxis": {
                    "range": [-1, 1],
                    "showgrid": False,
                    "showbackground": False,
                    "visible": False,
                },
                "zaxis": {
                    "range": [-1, 1],
                    "showgrid": False,
                    "showbackground": False,
                    "visible": False,
                },
                "aspectmode": "cube",
                "camera": {
                    "eye": {
                        "x": 0.8,
                        "y": 0.8,
                        "z": 0.8,
                    },
                },
            },
            showlegend=False,
        )

        return fig

    def show(
        self,
        format_str: str | None = None,
        mode: Literal["latex", "str"] | None = None,
    ) -> str | Math:
        r"""Return the quantum state as a string.

        Use the format string to change the print format of the basis states:

        * ``i``: print the state in the decimal base
        * ``b``: print the state in the binary base (default)
        * ``i|b<n>``: separate the ``n`` first qubits; the remaining print in the binary base
        * ``i|b<n1>:i|b<n2>[:i|b<n3>...]``: separate the ``n1, n2, n3, ...`` first qubits

        :Example:

            .. code-block:: py

                from ket import *

                p = Process()

                q = p.alloc(19)
                X(ctrl(H(q[0]), X, q[1:])[1::2])
                d = dump(q)

                print(d.show('i'))
                # |87381⟩ (50.00%)
                #  0.707107               ≅      1/√2
                # |436906⟩        (50.00%)
                #  0.707107               ≅      1/√2
                print(d.show('b'))
                # |0010101010101010101⟩   (50.00%)
                #  0.707107               ≅      1/√2
                # |1101010101010101010⟩   (50.00%)
                #  0.707107               ≅      1/√2
                print(d.show('i4'))
                # |2⟩|101010101010101⟩    (50.00%)
                #  0.707107               ≅      1/√2
                # |13⟩|010101010101010⟩   (50.00%)
                #  0.707107               ≅      1/√2
                print(d.show('b5:i4'))
                # |00101⟩|5⟩|0101010101⟩  (50.00%)
                #  0.707107               ≅      1/√2
                # |11010⟩|10⟩|1010101010⟩ (50.00%)
                #  0.707107               ≅      1/√2

        Args:
            format: Format string that matches ``(i|b)\d*(:(i|b)\d+)*``.
            mode: If ``"str"``, return a string representation of the quantum state. If ``"latex"``,
                return a LaTeX Math representation of the quantum state. If in Jupyter Notebook,
                defaults to ``"latex"``, otherwise defaults to ``"str"``.

        Returns:
            The formatted quantum state as a string, or Latex Math.
        """

        if mode is None:
            if IN_NOTEBOOK:
                mode = "latex"
            else:
                mode = "str"
        elif mode not in ("latex", "str"):
            raise ValueError(f"Unknown mode: {mode}")

        if format_str is not None:
            if format_str in ("b", "i"):
                format_str += str(self.size)
            fmt = []
            count = 0
            for b, size in map(lambda f: (f[0], int(f[1:])), format_str.split(":")):
                fmt.append((b, count, count + size))
                count += size
            if count < self.size:
                fmt.append(("b", count, self.size))
        else:
            fmt = [("b", 0, self.size)]

        if mode == "latex":
            return self._show_latex(fmt)
        return self._show_str(fmt)

    def _show_str(self, fmt=list[tuple[Literal["i", "b"], int, int]]) -> str:

        def fmt_ket(state, begin, end, f):
            return (
                f"|{state[begin:end]}⟩"
                if f == "b"
                else f"|{int(state[begin:end], base=2)}⟩"
            )

        def state_amp_str(state, amp):
            dump_str = "".join(
                fmt_ket(f"{state:0{self.size}b}", b, e, f) for f, b, e in fmt
            )
            dump_str += f"\t({100*abs(amp)**2:.2f}%)\n"
            real = abs(amp.real) > 1e-10
            real_l0 = amp.real < 0

            imag = abs(amp.imag) > 1e-10
            imag_l0 = amp.imag < 0

            sqrt_dem = 1 / abs(amp) ** 2
            use_sqrt = abs(round(sqrt_dem) - sqrt_dem) < 0.001
            use_sqrt = use_sqrt and (
                (abs(abs(amp.real) - abs(amp.imag)) < 1e-6) or (real != imag)
            )
            sqrt_dem = f"/√{round(1/abs(amp)**2)}"

            if real and imag:
                sqrt_dem = f"/√{round(2*(1/abs(amp)**2))}"
                sqrt_num = ("(-1" if real_l0 else " (1") + ("-i" if imag_l0 else "+i")
                sqrt_str = (
                    f"\t≅ {sqrt_num}){sqrt_dem}"
                    if use_sqrt and (abs(amp.real) - abs(amp.real) < 1e-10)
                    else ""
                )
                dump_str += f"{amp.real:9.6f}{amp.imag:+.6f}i" + sqrt_str
            elif real:
                sqrt_num = "  -1" if real_l0 else "   1"
                sqrt_str = f"\t≅   {sqrt_num}{sqrt_dem}" if use_sqrt else ""
                dump_str += f"{amp.real:9.6f}       " + sqrt_str
            else:
                sqrt_num = "  -i" if imag_l0 else "   i"
                sqrt_str = f"\t≅   {sqrt_num}{sqrt_dem}" if use_sqrt else ""
                dump_str += f" {amp.imag:17.6f}i" + sqrt_str

            return dump_str

        return "\n".join(
            state_amp_str(state, amp)
            for state, amp in sorted(self.get().items(), key=lambda k: k[0])
        )

    def _show_latex(self, fmt=list[tuple[Literal["i", "b"], int, int]]) -> Math:
        def float_to_math(num: float, is_complex: bool) -> str | None:
            num_str = None
            if abs(num) > 1e-14:

                sqrt_dem_float = 1 / num**2
                sqrt_dem = round(sqrt_dem_float)
                if abs(sqrt_dem - sqrt_dem_float) < 1e-10 and sqrt_dem != 1:
                    num_str = f"\\frac{{{'-' if num < 0.0 else ''}{'i' if is_complex else '1'}}}{{\\sqrt{{{sqrt_dem}}}}}"  # pylint: disable=line-too-long
                else:
                    round_num = round(num)
                    if abs(round_num - num) > 1e-14:
                        num_str = str(num)
                        if "e" in num_str:
                            num_str = num_str.replace("e", "\\times10^{") + "}"
                    else:
                        num_str = ""
                    if is_complex:
                        num_str += "i"
            return num_str

        math = []
        for state, amp in self.get().items():
            if abs(amp) < 1e-13:
                continue

            real_str = float_to_math(amp.real, False)
            imag_str = float_to_math(amp.imag, True)

            state_str = f"{state:0{len(self.qubits)}b}"
            state_str = [
                f"\\left|{state_str[start:end] if fmt == 'b' else int(state_str[start:end], 2)}\\right>"  # pylint: disable=line-too-long
                for fmt, start, end in fmt
            ]
            state_str = "".join(state_str)

            if real_str is not None and imag_str is not None:
                math.append(f"({real_str}+{imag_str}) {state_str}")
            else:
                math.append(
                    f"{real_str if real_str is not None else imag_str} {state_str}"
                )

        return Math("+".join(math).replace("+-", "-"))

    def histogram(self, mode: Literal["bin", "dec"] = "dec", **kwargs) -> go.Figure:
        """Generate a histogram representing the quantum state.

        This method creates a histogram visualizing the probability distribution
        of the quantum state.

        Note:
            This method requires additional dependencies from ``ket-lang[plot]``.

            Install with: ``pip install ket-lang[plot]``.

        Args:
            mode: If ``"bin"``, display the states in binary format. If ``"dec"``,
            display the states in decimal format. Defaults to ``"dec"``.
            **kwargs: Additional keyword arguments passed to :func:`plotly.express.bar`.

        Returns:
            Histogram of the quantum state.
        """
        _check_visualize()

        state = list(self.get().keys())
        state_text = (
            list(map(lambda s: f"|{s:0{self.size}b}⟩", state))
            if mode == "bin"
            else state
        )

        data = {
            "State": state,
            "Probability": list(map(lambda a: abs(a) ** 2, self.get().values())),
            "Phase": list(map(phase, self.get().values())),
        }

        fig = px.bar(
            data,
            x="State",
            y="Probability",
            color="Phase",
            range_color=(-pi, pi),
            **kwargs,
        )

        fig.update_layout(
            xaxis={
                "tickmode": "array",
                "ticktext": state_text,
                "tickvals": state,
            },
            bargap=0.75,
        )

        return fig

    def __repr__(self):
        return f"<Ket 'QuantumState' index={self.index}, pid={hex(id(self.process))}>"



# SPDX-FileCopyrightText: 2020 Evandro Chagas Ribeiro da Rosa <evandro@quantuloop.com>
# SPDX-FileCopyrightText: 2020 Rafael de Santiago <r.santiago@ufsc.br>
#
# SPDX-License-Identifier: Apache-2.0

"""Grover's search algorithm."""

from math import sqrt, pi
from typing import Callable
import ket
from random import randint


def grover(size: int, oracle: Callable, outcomes: int = 1) -> int:
    """Grover's search algorithm.

    Args:
        size (int): The number of qubits to use for the search.
        oracle (Callable): The oracle function to use for the search.
        outcomes (int): The number of outcomes to search for.

    Returns:
        int: The measured value from the search.

    """

    p = ket.Process(simulator="dense", num_qubits=size)

    s = ket.H(
        p.alloc(size)
    )  # Initialize the state to a superposition of all possible outcomes.

    steps = int(
        (pi / 4) * sqrt(2**size / outcomes)
    )  # Calculate the number of iterations for the algorithm.
    for _ in range(steps):
        oracle(s)  # Apply the oracle function to the state.
        with ket.around(ket.cat(ket.H, ket.X), s):
            ket.CZ(*s)  # Apply the diffusor function to the state.

    return ket.measure(s).value  # Measure the final state and return the result.


def test_grover():
    SIZE = 12
    looking_for = randint(0, pow(2, SIZE) - 1)
    assert looking_for == grover(SIZE, ket.qulib.oracle.phase_oracle(looking_for))


if __name__ == "__main__":
    test_grover()
    print("OK")



# SPDX-FileCopyrightText: 2024 Evandro Chagas Ribeiro da Rosa <evandro@quantuloop.com>
#
# SPDX-License-Identifier: Apache-2.0

from ket import X, Y, Z, H, T, S, RX, RY, RZ, C, Process, ctrl, using_aux, around, kron
from ket.qulib import dump_matrix
import numpy as np

GATES = {
    X: [[0, 1], [1, 0]],
    Y: [[0, -1j], [1j, 0]],
    Z: [[1, 0], [0, -1]],
    H: [[1 / np.sqrt(2), 1 / np.sqrt(2)], [1 / np.sqrt(2), -1 / np.sqrt(2)]],
    T: [[1, 0], [0, np.exp(1j * np.pi / 4)]],
    S: [[1, 0], [0, 1j]],
}

ROTATION_GATES = {
    RX: lambda theta: [
        [np.cos(theta / 2), -1j * np.sin(theta / 2)],
        [-1j * np.sin(theta / 2), np.cos(theta / 2)],
    ],
    RY: lambda theta: [
        [np.cos(theta / 2), -np.sin(theta / 2)],
        [np.sin(theta / 2), np.cos(theta / 2)],
    ],
    RZ: lambda theta: [[np.exp(-1j * theta / 2), 0], [0, np.exp(1j * theta / 2)]],
}


def validate(result, matrix):
    """Check resulting gate matrix"""

    gate_matrix = [
        [result[-2][-2], result[-2][-1]],
        [result[-1][-2], result[-1][-1]],
    ]

    eye_matrix = result

    eye_matrix[-2][-2] = 1
    eye_matrix[-2][-1] = 0
    eye_matrix[-1][-2] = 0
    eye_matrix[-1][-1] = 1

    return np.allclose(eye_matrix, np.eye(len(eye_matrix))) and np.allclose(
        gate_matrix, matrix
    )


def base_test_su2(c, a):
    """Test the decomposition of Rotation gates."""
    n = 2 * (c + 1) + a

    assert all(
        validate(
            dump_matrix(
                C(gate(angle)),
                num_qubits=(c, 1),
                process=Process(
                    num_qubits=n,
                    coupling_graph=[(i, j) for i in range(n) for j in range(i)],
                    simulator="sparse",
                ),
            ),
            matrix(angle),
        )
        for gate, matrix in ROTATION_GATES.items()
        for angle in np.linspace(0.0, 2 * np.pi, 8)
    )


def base_test_u2(c, a):
    """Test the decomposition of U2 gates."""
    n = 2 * (c + 1) + a

    assert all(
        validate(
            dump_matrix(
                C(gate),
                num_qubits=(c, 1),
                process=Process(
                    num_qubits=n,
                    coupling_graph=[(i, j) for i in range(n) for j in range(i)],
                    simulator="sparse",
                ),
            ),
            matrix,
        )
        for gate, matrix in GATES.items()
    )


def test_u2():
    """Test the decomposition of U2 gates up to 6 qubits."""
    n = 6
    for c in range(1, n):
        for a in range(c):
            base_test_u2(c, a)


def test_su2():
    """Test the decomposition of SU2 gates up to 6 qubits."""
    n = 6
    for c in range(1, n):
        for a in range(c):
            base_test_su2(c, a)


@using_aux(a=lambda c: 0 if len(c) <= 2 else 1)
def v_chain(c, t, a):
    """V Chain decomposition with clean auxiliary."""
    if len(c) <= 2:
        ctrl(c, X)(t)
    else:
        with around(ctrl(c[:2], X), a):
            v_chain(a + c[2:], t)


def test_v_chain():
    c = 6
    a = c - 2
    n = 2 * (c + 1) + a

    validate(
        dump_matrix(
            v_chain,
            num_qubits=(c, 1),
            process=Process(
                num_qubits=n,
                coupling_graph=[(i, j) for i in range(n) for j in range(i)],
                simulator="sparse",
            ),
        ),
        GATES[X],
    )


@using_aux(a=lambda c: len(c) // 2 if len(c) > 2 else 0)
def network(gate, c, t, a):
    n = len(c)

    if n <= 2:
        ctrl(c, gate)(t)
    else:
        num_groups = n // 2
        rest = n % 2

        with around(
            kron(C(X), n=num_groups),
            *((c[2 * i : 2 * i + 2], a[i]) for i in range(num_groups))
        ):
            network(
                gate,
                c[-rest:] + a if rest != 0 else a,
                t,
            )


def test_network_h():
    c = 6
    a = c - 2
    n = 2 * (c + 1) + a

    validate(
        dump_matrix(
            network,
            num_qubits=(c, 1),
            args=(H,),
            process=Process(
                num_qubits=n,
                coupling_graph=[(i, j) for i in range(n) for j in range(i)],
                simulator="sparse",
            ),
        ),
        GATES[H],
    )


@using_aux(unsafe=True, a=lambda c: len(c) - 2)
def v_chain_dirty(c, t, a):
    def inner(c, t, a, skip=False):
        if len(c) <= 2:
            ctrl(c, X)(t)
        elif skip:
            inner(c=c[:-1], t=a[-1], a=a[:-1])
        else:
            with around(ctrl(c[-1] + a[-1], X), t):
                inner(c=c[:-1], t=a[-1], a=a[:-1])

    inner(c, t, a)
    inner(c, t, a, True)


def test_v_chain_dirty():
    c = 6
    a = c - 2
    n = 2 * (c + 1) + a

    validate(
        dump_matrix(
            v_chain_dirty,
            num_qubits=(c, 1),
            process=Process(
                num_qubits=n,
                coupling_graph=[(i, j) for i in range(n) for j in range(i)],
                simulator="sparse",
            ),
        ),
        GATES[X],
    )


if __name__ == "__main__":
    test_su2()
    test_u2()
    test_v_chain()
    test_network_h()
    test_v_chain_dirty()
    print("OK")



"""Quantum Gate Construction."""

from __future__ import annotations

# SPDX-FileCopyrightText: 2025 Evandro Chagas Ribeiro da Rosa <evandro@quantuloop.com>
#
# SPDX-License-Identifier: Apache-2.0

from functools import partial
from itertools import starmap
from typing import Callable
from cmath import isclose, phase as cmath_phase, exp
from math import atan2
from ..gates import X, P, RZ, RY, CNOT, global_phase
from ..operations import adj, around
from ..base import Quant

__all__ = [
    "unitary",
    "diagonal",
    "ucz",
]


def _is_unitary(matrix):
    if len(matrix) != 2 or len(matrix[0]) != 2 or len(matrix[1]) != 2:
        raise ValueError("Input matrix must be a 2x2 matrix")

    conj_transpose = [[matrix[j][i].conjugate() for j in range(2)] for i in range(2)]

    result = [
        [sum(matrix[i][k] * conj_transpose[k][j] for k in range(2)) for j in range(2)]
        for i in range(2)
    ]

    return all(
        isclose(result[i][j], 1 if i == j else 0, abs_tol=1e-10)
        for i in range(2)
        for j in range(2)
    )


def _zyz(matrix):
    a, b = matrix[0]
    c, d = matrix[1]

    det = cmath_phase(a * d - b * c)
    phase = det / 2

    theta = 2 * atan2(abs(c), abs(a))

    ang1 = cmath_phase(d)
    ang2 = cmath_phase(c)

    phi = ang1 + ang2 - det
    lam = ang1 - ang2

    return phase, phi, theta, lam


def _apply_global_phase(phase: float, qubits: Quant):
    P(phase, qubits[0])
    with around(X, qubits[0]):
        P(phase, qubits[0])


def unitary(
    matrix: list[list[complex]], up_to_global_phase: bool = False
) -> Callable[[Quant], Quant]:
    """Create a quantum gate from 2x2 unitary matrix.

    The provided unitary matrix is decomposed into a sequence of rotation gates,
    which together implement an equivalent unitary transformation. When the gate
    is used in a controlled operation, the resulting unitary is equivalent up to
    a global phase.

    Args:
        matrix: Unitary matrix in the format ``[[a, b], [c, d]]``.
        up_to_global_phase: If True, the resulting unitary is equivalent up to
            a global phase even when the gate is not used in a controlled
            operation.

    Returns:
        Returns a new callable that implements the unitary operation.

    Raises:
        ValueError: If the input matrix is not unitary.
    """
    if not _is_unitary(matrix):
        raise ValueError("Input matrix is not unitary")

    phase, phi, theta, lam = _zyz(matrix)

    @global_phase(0.0 if up_to_global_phase else phase)
    def inner(qubits: Quant):
        if up_to_global_phase:
            _apply_global_phase(phase, qubits)
        RZ(lam.real, qubits)
        RY(theta.real, qubits)
        RZ(phi.real, qubits)
        return qubits

    return inner


def ucz(angles, qubits, *, _last=True):
    """Uniformly controlled z-axis rotation gate.

    Implements a uniformly controlled rotation around the z-axis, as proposed
    in https://arxiv.org/abs/quant-ph/0406176.

    Args:
        angles: List of rotation angles.
        qubits: List of qubits to apply the gate to.
    """
    if len(angles) != 1 << (len(qubits) - 1):
        raise ValueError(
            f"Invalid number of angles. Expected {1 << (len(qubits) - 1)}, got {len(angles)}."
        )
    if len(qubits) == 1:
        RZ(angles[0], qubits)
        return

    h = len(angles) // 2
    uh = angles[:h]
    lh = angles[h:]

    angles1 = list(starmap(lambda a, b: (a + b) / 2, zip(lh, uh)))
    angles2 = list(starmap(lambda a, b: (a - b) / 2, zip(lh, uh)))

    ucz(angles1, qubits[1:], _last=False)
    CNOT(qubits[0], qubits[-1])
    adj(ucz)(angles2, qubits[1:], _last=False)
    if _last:
        CNOT(qubits[0], qubits[-1])


def _diagonal_gate(angles, qubits, up_to_global_phase=True):
    if len(angles) == 1:
        return angles[0]

    n = len(angles)

    rz_angle = [
        (exp(1j * angles[2 * i]), exp(1j * angles[2 * i + 1])) for i in range(n // 2)
    ]

    rz_fix = [cmath_phase(a[0] * a[1]) / 2 for a in rz_angle]

    rz_angle = list(starmap(lambda r, a: r[0] * exp(-a * 1j), zip(rz_angle, rz_fix)))

    rz_angle = [-cmath_phase(a) * 2 for a in rz_angle]

    ucz(rz_angle, qubits)

    phase = _diagonal_gate(rz_fix, qubits[:-1], False)

    if up_to_global_phase:
        _apply_global_phase(phase, qubits)
    return phase


def diagonal(*angles, up_to_global_phase=True):
    r"""Create a diagonal quantum gate.

    .. math::

        U = \sum_i \theta_i \left| i \right> \left< i \right|

    Args:
        angles: List of rotation angles :math:`\theta_i`.
        up_to_global_phase: If True, the resulting unitary is equivalent up to
            a global phase.
    """

    if len(angles).bit_count() != 1:
        raise ValueError("Invalid number of angles. Expected a power of 2.")

    return partial(_diagonal_gate, angles, up_to_global_phase=up_to_global_phase)



# SPDX-FileCopyrightText: 2020 Evandro Chagas Ribeiro da Rosa <evandro@quantuloop.com>
# SPDX-FileCopyrightText: 2020 Rafael de Santiago <r.santiago@ufsc.br>
#
# SPDX-License-Identifier: Apache-2.0
"""Ket Quantum Programming Platform.

Ket is a comprehensive library designed for quantum programming, providing essential functions and
classes to build quantum algorithms and applications. It facilitates the manipulation and storage of
quantum states, measurement operations, and the computation of expected values.

Explore the documentation of individual submodules for in-depth information and more practical code
examples.

All the functionality from the submodules is conveniently accessible within the ``ket`` namespace.
Except for the `qulib` module.

Examples:

    - Grover algorithm

    .. code-block:: python

        from math import sqrt, pi
        import ket

        def grover(size: int, oracle) -> int:
            p = ket.Process(simulator="dense", num_qubits=size)

            s = ket.H(p.alloc(size))

            steps = int((pi / 4) * sqrt(2**size))

            for _ in range(steps):
                oracle(s)
                with ket.around(ket.H, s):
                    ket.phase_oracle(0, s)

            return ket.measure(s).value

    - Quantum teleportation protocol

    .. code-block:: python

        import ket

        def teleport(alice_msg, alice_aux, bob_aux):
            ket.ctrl(alice_msg, ket.X)(alice_aux)
            ket.H(alice_msg)

            m0 = ket.measure(alice_msg)
            m1 = ket.measure(alice_aux)

            if m1.value == 1:
                ket.X(bob_aux)
            if m0.value == 1:
                ket.Z(bob_aux)

            return bob_aux

        def bell(qubits):
            return ket.ctrl(ket.H(qubits[0]), ket.X)(qubits[1])

        def message(qubit):
            ket.H(alice)
            ket.Z(alice)

        p = ket.Process()

        alice = p.alloc()  # alice = |0⟩
        message(alice)     # alice = |–⟩

        bob = teleport(alice, *bell(p.alloc(2)))  # bob  <- alice

        ket.H(bob)         # bob   = |1⟩
        bob_m = ket.measure(bob)
        print("Expected measure 1, result =", bob_m.value)

"""


from .clib import libs
from .base import *
from .base import __all__ as all_base
from .operations import *
from .operations import __all__ as all_func
from .gates import *
from .gates import __all__ as all_gate
from .expv import *
from .expv import __all__ as all_expv
from .quantumstate import *
from .quantumstate import __all__ as all_state
from . import qulib

__version__ = "0.9.1rc3"

__all__ = all_base + all_func + all_gate + all_expv + all_state + ["qulib"]


def ket_version() -> list[str]:
    """Return the version of the Ket platform components."""
    from .clib.libket import API as libket  # pylint: disable=import-outside-toplevel
    from .clib.kbw import API as kbw  # pylint: disable=import-outside-toplevel

    libket_v, size = libket["ket_build_info"]()
    libket_v = bytearray(libket_v[: size.value])
    libket_v = libket_v.decode("utf-8")

    kbw_v, size = kbw["kbw_build_info"]()
    kbw_v = bytearray(kbw_v[: size.value])
    kbw_v = kbw_v.decode("utf-8")

    return [f"Ket v{__version__}", libket_v, kbw_v]



"""Remote quantum execution module.

This module provides a tool to execute quantum code on a remote server. It includes
the :class:`~ket.remote.Remote` class, which facilitates the connection to a remote
server.
"""

from __future__ import annotations

# SPDX-FileCopyrightText: 2024 Evandro Chagas Ribeiro da Rosa <evandro@quantuloop.com>
#
# SPDX-License-Identifier: Apache-2.0


from .clib.libket import BatchExecution

try:
    import requests

    REQUESTS_AVAILABLE = True
except ImportError:
    requests = None
    REQUESTS_AVAILABLE = False

try:
    import jwt
    from cryptography.hazmat.primitives import serialization
    import getpass
    from os import PathLike, path
    from datetime import datetime, timezone, timedelta

    JWT_AVAILABLE = True
except ImportError:
    JWT_AVAILABLE = False


class Remote(
    BatchExecution
):  # pylint: disable=too-few-public-methods,too-many-instance-attributes
    """Remote quantum execution.

    The :class:`~ket.remote.Remote` class enables the execution of quantum circuits
    on a remote server. It provides functionality to establish a connection with
    the server and configure the connection parameters.

    The remote server's configuration is specified using the ``url`` parameter, while
    additional connection parameters can be passed as keyword arguments to the
    :meth:`~ket.remote.Remote.connect` method. The exact arguments required depend
    on the server's implementation. Refer to the server's documentation for detailed
    setup instructions.

    Once connected, the :class:`~ket.remote.Remote` instance integrates with a
    :class:`~ket.base.Process` object, which handles the actual quantum execution.
    The server determines the capabilities available for execution, such as
    calculating expectation values or dumping quantum states. The execution is set to
    batch.

    .. note::

        This module requires additional dependencies from ``ket-lang[remote]``.

        Install with: ``pip install ket-lang[remote]``.

    Example:

        .. code-block:: python

            from ket.remote import Remote

            # Initialize a remote server connection
            remote = Remote("https://remote-server-url.com")

            # Connect to the remote server and start a quantum process
            process = Process(remote.connect(...))


    Args:
        url: The URL of the remote server.
        timeout: The timeout for the requests. Defaults to None.
        verify_ssl: Flag to disable SSL certificate verification.
            This value is passed directly to the ``requests.get()`` call.
            Defaults to :obj:`None`.
        private_key: Path to an OpenSSH RSA private key for authentication.
            If set to :obj:`None`, authentication will not be performed. Defaults to :obj:`None`.
        passphrase: Passphrase to decrypt the private key:

            - :obj:`True`: Prompts the user to enter the passphrase interactively;
            - :class:`bytes`: Provides the passphrase directly as a byte string;
            - :obj:`None`: Assumes the private key is unencrypted. Defaults to :obj:`None`.
    """

    def __init__(  # pylint: disable=too-many-arguments,too-many-positional-arguments
        self,
        url: str,
        verify_ssl=None,
        timeout: int | None = None,
        private_key: PathLike | None = None,
        passphrase: bool | bytes | None = None,
    ):
        super().__init__()
        if not REQUESTS_AVAILABLE:
            raise ImportError("requests is not installed, run `pip install requests`")
        self._url = url
        self._result = None
        self._timeout = timeout
        self._result_json = bytearray()
        self._result_len = 0
        self._args = None
        self._verify_ssl = verify_ssl

        if private_key is not None:
            if not JWT_AVAILABLE:
                raise ImportError(
                    "jwt is not installed, run `pip install PyJWT[crypto] bcrypt`"
                )

            if isinstance(passphrase, bool):
                if passphrase:
                    passphrase = bytes(
                        getpass.getpass(f"passphrase for {private_key}:"), "utf-8"
                    )
                else:
                    passphrase = None

            with open(path.expandvars(path.expanduser(private_key)), "rb") as key_file:
                private_key = serialization.load_ssh_private_key(
                    key_file.read(), password=passphrase
                )
            self._token = jwt.encode(
                {
                    "iss": "org.quantumket.ket",
                    "exp": datetime.now(tz=timezone.utc)
                    + timedelta(seconds=timeout if timeout is not None else 604800),
                },
                private_key,
                algorithm="PS256",
            )
        else:
            self._token = None

    def submit_execution(
        self,
        circuit: dict,
        parameters: list[float],
    ):
        url = f"{self._url}/run"
        response = requests.get(
            url,
            verify=self._verify_ssl,
            json=(circuit, parameters, self._args),
            timeout=self._timeout,
        )
        if response.status_code == 200:
            self._result = response.json()
        else:
            raise RuntimeError(f"Error: {response.status_code} - {response.text}")

    def get_result(self) -> dict:
        return self._result

    def clear(self):
        pass

    def set(self, **kwargs):
        """Configure the execution.

        Args:
            kwargs: Keyword arguments specifying connection parameters. The required
                    arguments depend on the remote server's API.
        """
        self._args = {k: str(v) for k, v in kwargs.items()}
        return self

    def connect(
        self,
    ):
        if self._token is not None:
            self._args["token"] = self._token

        url = f"{self._url}/get"
        response = requests.get(
            url,
            verify=self._verify_ssl,
            json=self._args,
            timeout=self._timeout,
        )
        if response.status_code == 200:
            result = response.json()
        else:
            raise RuntimeError(f"Error: {response.status_code} - {response.text}")

        return super().configure(**result)



# SPDX-FileCopyrightText: 2020 Evandro Chagas Ribeiro da Rosa <evandro@quantuloop.com>
# SPDX-FileCopyrightText: 2020 Rafael de Santiago <r.santiago@ufsc.br>
#
# SPDX-License-Identifier: Apache-2.0

"""C Libs"""

import os
import shutil
import warnings


def libs():
    """Name Rust libs"""
    if os.name == "nt":
        return [("libket", "ket.dll"), ("kbw", "kbw.dll")]
    if os.uname().sysname == "Linux":
        return [("libket", "libket.so"), ("kbw", "libkbw.so")]
    if os.uname().sysname == "Darwin":
        return [("libket", "libket.dylib"), ("kbw", "libkbw.dylib")]
    raise OSError("unsupported operational system")


def make_libs():
    """Make Rust libs if missing"""
    dirname = os.path.dirname(__file__)

    if all(os.path.isfile(f"{dirname}/{lib}") for _, lib in libs()):
        return

    warnings.warn(
        "Compiling Libket and KBW... may take a while. \
If it fails, check if Rust is installed https://www.rust-lang.org/tools/install."
    )

    os.chdir(f"{dirname}/libket")
    os.system("cargo build --release --quiet")
    os.chdir("../kbw")
    os.system("cargo build --release --quiet")

    for dirlib, lib in libs():
        src = f"{dirname}/{dirlib}/target/release/{lib}"
        dst = f"{dirname}/{lib}"
        shutil.move(src, dst)


make_libs()



"""Live Execution Simulator Based on NumPy

This simulator is available as a demonstration of the LiveExecution class for
constructing live execution targets using Python. It is not intended to be a
high-performance simulator. For better performance, use the KBW simulator.
"""

# SPDX-FileCopyrightText: 2025 Evandro Chagas Ribeiro da Rosa <evandro@quantuloop.com>
#
# SPDX-License-Identifier: Apache-2.0

from random import choices
from .clib.libket import LiveExecution

try:
    import numpy as np

    NP_AVAILABLE = True
except ImportError:
    NP_AVAILABLE = False


class NPSim(LiveExecution):
    """NumPy based Live Execution simulator.

    Args:
        num_qubits: number of qubits.
    """

    def __init__(self, num_qubits):
        if not NP_AVAILABLE:
            raise RuntimeError(
                "NumPy is not available. Please install it with: pip install numpy"
            )

        super().__init__()

        self.num_qubit = num_qubits
        self.state_vector = np.zeros((1 << num_qubits, 1), dtype=complex)
        self.state_vector[0][0] = 1

    def connect(self):
        return self.configure(
            num_qubits=self.num_qubit,
            execution_managed_by_target={
                "measure": "Advanced",
                "dump": "Advanced",
            },
            qpu={"u4_gate": "CX"},
        )

    def _apply_gate(self, gate, target):
        gate = np.kron(
            np.kron(np.eye(1 << target, 1 << target, dtype=complex), gate),
            np.eye(
                1 << (self.num_qubit - target - 1),
                1 << (self.num_qubit - target - 1),
                dtype=complex,
            ),
        )
        self.state_vector = gate @ self.state_vector

    def pauli_x(self, target, control):
        """Apply a Pauli-X gate to the target qubit."""
        assert len(control) <= 1, "Multi-control qubits are not supported"
        gate = np.array([[0, 1], [1, 0]], dtype=complex)
        if len(control) == 0:
            self._apply_gate(gate, target)
        else:
            gate = np.zeros((1 << self.num_qubit, 1 << self.num_qubit), dtype=complex)
            for col in map(
                lambda s: list(f"{s:0{self.num_qubit}b}"), range(1 << self.num_qubit)
            ):
                lin = col[:]
                if col[control[0]] == "1":
                    lin[target] = "0" if col[target] == "1" else "1"
                gate[int("".join(lin), 2)][int("".join(col), 2)] = 1.0
            self.state_vector = gate @ self.state_vector

    def pauli_y(self, target, control):
        """Apply a Pauli-Y gate to the target qubit."""
        assert len(control) != 0, "control qubits are not supported"
        gate = np.array([[0, -1j], [1j, 0]], dtype=complex)
        self._apply_gate(gate, target)

    def pauli_z(self, target, control):
        """Apply a Pauli-Z gate to the target qubit."""
        assert len(control) != 0, "control qubits are not supported"
        gate = np.array([[1, 0], [0, -1]], dtype=complex)
        self._apply_gate(gate, target)

    def hadamard(self, target, control):
        """Apply a Hadamard gate to the target qubit."""
        assert len(control) == 0, "Control qubits are not supported"
        gate = np.array([[1, 1], [1, -1]], dtype=complex) / np.sqrt(2)
        self._apply_gate(gate, target)

    def rotation_x(self, target, control, angle):
        """Apply a X-Rotation gate to the target qubit."""
        assert len(control) == 0, "Control qubits are not supported"
        gate = np.array(
            [
                [np.cos(angle / 2), -1j * np.sin(angle / 2)],
                [-1j * np.sin(angle / 2), np.cos(angle / 2)],
            ],
            dtype=complex,
        )
        self._apply_gate(gate, target)

    def rotation_y(self, target, control, angle):
        """Apply a Y-Rotation gate to the target qubit."""
        assert len(control) == 0, "Control qubits are not supported"
        gate = np.array(
            [
                [np.cos(angle / 2), -1j * np.sin(angle / 2)],
                [np.sin(angle / 2), np.cos(angle / 2)],
            ],
            dtype=complex,
        )
        self._apply_gate(gate, target)

    def rotation_z(self, target, control, angle):
        """Apply a Z-Rotation gate to the target qubit."""
        assert len(control) == 0, "Control qubits are not supported"
        gate = np.array(
            [[np.exp(-1j * angle / 2), 0], [0, np.exp(1j * angle / 2)]], dtype=complex
        )
        self._apply_gate(gate, target)

    def phase(self, target, control, angle):
        """Apply a Phase gate to the target qubit."""
        assert len(control) == 0, "Control qubits are not supported"
        gate = np.array([[1, 0], [0, np.exp(1j * angle)]], dtype=complex)
        self._apply_gate(gate, target)

    def _p1(self, qubit: int) -> float:
        return sum(
            np.abs(amp) ** 2 if state & (1 << qubit) != 0 else 0.0
            for state, amp in enumerate(self.state_vector)
        )

    def _colapse(self, qubit, m, p):
        for state, amp in enumerate(self.state_vector):
            if state & (1 << qubit) == m:
                amp /= np.sqrt(p)
            else:
                self.state_vector[state] = 0.0

    def measure(self, qubits: list[int]) -> int:
        result = []
        for q in reversed(qubits):
            p1 = self._p1(q)
            p0 = 1.0 - p1
            (m, p) = choices([(0, p0), (1, p1)], [p0, p1])[0]
            result.append(m)
            self._colapse(q, m, p)
        return int("".join(map(str, result)), 2)

    def dump(self, qubits: list[int]) -> dict:
        basis_states = []
        amplitudes_real = []
        amplitudes_imag = []

        for state, amp in enumerate(self.state_vector):
            if np.abs(amp) <= 1e-10:
                continue
            state = f"{state:0{self.num_qubit}b}"
            state = int("".join(state[q] for q in qubits), 2)
            basis_states.append([state])
            amplitudes_real.append(float(np.real(amp)))
            amplitudes_imag.append(float(np.imag(amp)))

        return {
            "basis_states": basis_states,
            "amplitudes_real": amplitudes_real,
            "amplitudes_imag": amplitudes_imag,
        }



"""Module providing functionality to interact with IBM Quantum and IBM Cloud devices.

Note:
    This module requires additional dependencies from ``ket-lang[ibm]``.

    Install with: ``pip install ket-lang[ibm]``.

"""

from __future__ import annotations

# SPDX-FileCopyrightText: 2024 Evandro Chagas Ribeiro da Rosa <evandro@quantuloop.com>
# SPDX-FileCopyrightText: 2024 Otávio Augusto de Santana Jatobá
# <otavio.jatoba@grad.ufsc.br>
#
# SPDX-License-Identifier: Apache-2.0

from .clib.libket import BatchExecution

try:
    from qiskit import QuantumCircuit
    from qiskit.circuit import library
    from qiskit.providers import BackendV2
    from qiskit.transpiler.preset_passmanagers import generate_preset_pass_manager
    from qiskit_ibm_runtime import SamplerV2 as Sampler
    from qiskit_aer import AerSimulator

    QISKIT_AVAILABLE = True
except ImportError:
    QISKIT_AVAILABLE = False


class IBMDevice(BatchExecution):  # pylint: disable=too-many-instance-attributes
    """IBM Qiskit backend for Ket process.

    The arguments ``shots`` and ``classical_shadows`` control how the
    execution is performed for estimating expectation values of an
    Hamiltonian term. Only one of these arguments can be specified at a time.

    If ``shots`` is specified, it will run the circuit multiple times
    (the number of shots) to estimate the expectation values.
    If ``classical_shadows`` is specified, it will use the classical shadows
    technique for state estimation. The dictionary should be in the
    format: ``{"bias": (int, int, int), "samples": int, "shots": int}``.
    The ``bias`` tuple represents the bias for the randomized measurements on the
    X, Y, and Z axes, respectively. The ``samples`` is the number of
    classical shadows to be generated, and ``shots`` is the number of shots
    for each sample.

    Args:
        backend: The backend to be used for the quantum execution. If not
            provided, it defaults to the AerSimulator.
        use_qiskit_transpiler: Use Qiskit transpiler instead of Ket's.
        shots: The number of shots for the execution to estimate
            the expectation values of an Hamiltonian term. If ``classical_shadows``
            and ``shots`` are not specified, it defaults to 2048.
        classical_shadows: If specified, it will use the classical
            shadows technique for state estimation.
    """

    def __init__(
        self,
        backend: BackendV2 | None = None,
        *,
        shots: int | None = None,
        classical_shadows: dict | None = None,
        use_qiskit_transpiler: bool = False,
    ):
        if not QISKIT_AVAILABLE:
            raise ImportError(
                "IBMDevice requires the qiskit module to be used. You can install them"
                " alongside ket by running `pip install ket[ibm]`."
            )

        if shots is not None and classical_shadows is not None:
            raise ValueError(
                "You cannot specify both 'shots' and 'classical_shadows'. Please"
                " choose one of them."
            )

        super().__init__()

        if backend is None:
            backend = AerSimulator()

        self.num_qubits = backend.configuration().n_qubits

        self.client = None
        self.backend = backend

        if not use_qiskit_transpiler:
            self.coupling_graph = (
                list(backend.coupling_map.graph.edge_list())
                if backend.coupling_map
                else [
                    [i, j]
                    for i in range(self.num_qubits)
                    for j in range(self.num_qubits)
                    if i != j
                ]
            )
        else:
            self.coupling_graph = None

        self.circuit = QuantumCircuit(self.num_qubits)
        self.parameters = None

        self.shots = 2048 if shots is None and classical_shadows is None else shots
        self.classical_shadows = classical_shadows

        self.qubits_from_sample = None
        self.result = None

    def clear(self):
        self.circuit = QuantumCircuit(self.num_qubits)
        self.parameters = None
        self.qubits_from_sample = None
        self.result = None

    def submit_execution(self, circuit, parameters):
        self.parameters = parameters
        self.process_instructions(circuit)

    def get_result(self):
        self.result = {
            "".join(s[-q - 1] for q in self.qubits_from_sample): c
            for s, c in self.result.items()
        }

        result = list(zip(*((int(s, 2), c) for s, c in self.result.items())))

        results_dict = {
            "measurements": [],
            "exp_values": [],
            "samples": [result],
            "dumps": [],
            "gradients": None,
        }

        return results_dict

    def pauli_x(self, target, control):
        gate = library.XGate()
        if len(control) >= 1:
            gate = gate.control(len(control))
        self.circuit.append(gate, control + [target])

    def pauli_y(self, target, control):
        gate = library.YGate()
        if len(control) >= 1:
            gate = gate.control(len(control))
        self.circuit.append(gate, control + [target])

    def pauli_z(self, target, control):
        gate = library.ZGate()
        if len(control) >= 1:
            gate = gate.control(len(control))
        self.circuit.append(gate, control + [target])

    def hadamard(self, target, control):
        gate = library.HGate()
        if len(control) >= 1:
            gate = gate.control(len(control))
        self.circuit.append(gate, control + [target])

    def rotation_x(self, target, control, **kwargs):
        gate = library.RXGate(kwargs["Value"])
        if len(control) >= 1:
            gate = gate.control(len(control))
        self.circuit.append(gate, control + [target])

    def rotation_y(self, target, control, **kwargs):
        gate = library.RYGate(
            kwargs["Value"]
            if "Value" in kwargs
            else self.parameters[kwargs["Ref"]["index"]] * kwargs["Ref"]["multiplier"]
        )
        if len(control) >= 1:
            gate = gate.control(len(control))
        self.circuit.append(gate, control + [target])

    def rotation_z(self, target, control, **kwargs):
        gate = library.RZGate(
            kwargs["Value"]
            if "Value" in kwargs
            else self.parameters[kwargs["Ref"]["index"]] * kwargs["Ref"]["multiplier"]
        )
        if len(control) >= 1:
            gate = gate.control(len(control))
        self.circuit.append(gate, control + [target])

    def phase(self, target, control, **kwargs):
        gate = library.PhaseGate(
            kwargs["Value"]
            if "Value" in kwargs
            else self.parameters[kwargs["Ref"]["index"]] * kwargs["Ref"]["multiplier"]
        )
        if len(control) >= 1:
            gate = gate.control(len(control))
        self.circuit.append(gate, control + [target])

    def sample(self, _, qubits, shots):
        self.circuit.measure_all()
        pm = generate_preset_pass_manager(
            backend=self.backend,
            initial_layout=(
                list(range(self.num_qubits))
                if self.coupling_graph is not None
                else None
            ),
        )
        isa_circuit = pm.run(self.circuit)
        sampler = Sampler(mode=self.backend)
        job = sampler.run([isa_circuit], shots=shots)
        self.result = job.result()[0].data.meas.get_counts()
        self.qubits_from_sample = qubits

    def connect(self):
        self.clear()

        qpu_params = {
            "coupling_graph": self.coupling_graph,
            "u4_gate_type": "CX",
            "u2_gate_set": "All",
        }

        return super().configure(
            num_qubits=self.num_qubits,
            direct_sample_exp_value=self.shots,
            classical_shadows_exp_value=self.classical_shadows,
            qpu=qpu_params if self.coupling_graph is not None else None,
        )



# SPDX-FileCopyrightText: 2020 Evandro Chagas Ribeiro da Rosa <evandro@quantuloop.com>
#
# SPDX-License-Identifier: Apache-2.0

from itertools import product
import math
import cmath
import ket
from ket import qulib


def u3_matrix(theta: float, phi: float, lambda_: float) -> list:
    return [
        [
            cmath.exp(-1j * (phi + lambda_) / 2) * math.cos(theta / 2),
            -cmath.exp(-1j * (phi - lambda_) / 2) * math.sin(theta / 2),
        ],
        [
            cmath.exp(1j * (phi - lambda_) / 2) * math.sin(theta / 2),
            cmath.exp(1j * (phi + lambda_) / 2) * math.cos(theta / 2),
        ],
    ]


def linspace(start, stop, num):
    yield start
    step = (stop - start) / (num - 1)
    next_num = start
    for _ in range(num - 1):
        next_num += step
        yield next_num


def test_u3_0_2pi_gate():
    for theta, phi, lambda_ in product(linspace(0, 2 * math.pi, 10), repeat=3):
        gate = qulib.dump_matrix(ket.U3(theta, phi, lambda_))
        matrix = u3_matrix(theta, phi, lambda_)

        assert all(
            cmath.isclose(gate[i][j], matrix[i][j], abs_tol=1e-7)
            for i in range(2)
            for j in range(2)
        )


def test_u3_m2pi_0_gate():
    for theta, phi, lambda_ in product(linspace(-2 * math.pi, 0.0, 10), repeat=3):
        gate = qulib.dump_matrix(lambda q: ket.U3(theta, phi, lambda_, q))
        matrix = u3_matrix(theta, phi, lambda_)

        assert all(
            cmath.isclose(gate[i][j], matrix[i][j], abs_tol=1e-7)
            for i in range(2)
            for j in range(2)
        )


if __name__ == "__main__":
    test_u3_0_2pi_gate()
    test_u3_m2pi_0_gate()
    print("OK")



"""Functions to manipulate quantum operations.

This module provides essential functions in the Ket library for manipulating quantum operations.
It includes functions for controlled and inverse operations, facilitating quantum circuit
construction.
"""

from __future__ import annotations

# SPDX-FileCopyrightText: 2020 Evandro Chagas Ribeiro da Rosa <evandro@quantuloop.com>
# SPDX-FileCopyrightText: 2020 Rafael de Santiago <r.santiago@ufsc.br>
#
# SPDX-License-Identifier: Apache-2.0


from collections.abc import Sized
from contextlib import contextmanager
from ctypes import c_size_t
from functools import reduce, wraps
from operator import add
from typing import Any, Callable, Sequence
from inspect import signature

from ket.clib.libket import PAULI_X


from .base import (
    Process,
    Quant,
    Measurement,
    Samples,
)
from .quantumstate import QuantumState


from .expv import (
    Pauli,
    Hamiltonian,
    ExpValue,
)

__all__ = [
    "control",
    "ctrl",
    "inverse",
    "adj",
    "around",
    "cat",
    "kron",
    "measure",
    "sample",
    "dump",
    "exp_value",
    "using_aux",
    "is_permutation",
    "is_diagonal",
    "C",
]


def _flip_to_control(
    control_state: int | list[int] | None, qubits: Quant | None = None
) -> Quant | Callable[[Quant], Quant]:
    r"""Flip qubits from :math:`\ket{\texttt{control_state}}` to :math:`\ket{1\dots1}`.

    The primary usage of this gate is to change the state when controlled applications are applied.
    For instance, all controlled operations are only applied if the control qubits' state is
    :math:`\ket{1}`. This gate is useful for using another state as control.

    Example:

        .. code-block:: python

            from ket import *

            p = Process()
            q = p.alloc(3)

            H(q[:2])

            with around(_flip_to_control(0b01), q[:2]):
                ctrl(q[:2], X)(q[2])

    Args:
        control_state: The state to flip the control qubits to.
        qubits: The qubits to apply the flip to.
    """

    def inner(qubits: Quant) -> Quant:
        if control_state is None:
            return qubits

        if not isinstance(qubits, Quant):
            qubits = reduce(add, qubits)

        length = len(qubits)
        if isinstance(control_state, Sized):
            if len(control_state) != length:
                raise ValueError(
                    f"'to' received a list of length {len(control_state)} to use on {length} qubits"
                )
            state = control_state
        else:
            if length < control_state.bit_length():
                raise ValueError(
                    f"To flip with control_state={control_state} "
                    f"you need at least {control_state.bit_length()} qubits"
                )

            state = [int(i) for i in f"{{:0{length}b}}".format(control_state)]

        for i, qubit in zip(state, qubits):
            if i == 0:
                qubit.process.apply_gate(PAULI_X, 0.0, False, 0, qubit.qubits[0])
        return qubits

    if qubits is None:
        return inner
    return inner(qubits)


@contextmanager
def control(control_qubits: Quant, state: int | list[int] | None = None):
    r"""Controlled scope.

    Opens a controlled scope where quantum operations are applied only if the qubits of the
    parameter ``control_qubits`` are in the state :math:`\ket{1}`. The parameter ``state`` can
    be used to specify a different state.

    :Usage:

    .. code-block:: python

        with control(control_qubits):
            # Apply operations only if control_qubits are in the state |1>
            ...

    Example:

        .. code-block:: python

            from ket import *

            p = Process()

            c = p.alloc(2)
            a, b = q.alloc(2)

            # CNOT c[0] a
            with control(c[0]):
                X(a)

            # Toffoli c[0] c[1] a
            with control(c):
                X(a)

            # CSWAP c[0] a b
            with control(c[0]):
                SWAP(a, b)

    Args:
        control_qubits: The qubits to control the quantum operations.
        state: The state to apply to the control qubits. Defaults to :math:`\ket{1}`.
    """

    if not isinstance(control_qubits, Quant):
        control_qubits = reduce(add, control_qubits)
    process = control_qubits.process

    if state is not None:
        with around(_flip_to_control(state), control_qubits, ket_process=process):
            process.ctrl_push(
                (c_size_t * len(control_qubits.qubits))(*control_qubits.qubits),
                len(control_qubits.qubits),
            )
            try:
                yield
            finally:
                process.ctrl_pop()
    else:
        process.ctrl_push(
            (c_size_t * len(control_qubits.qubits))(*control_qubits.qubits),
            len(control_qubits.qubits),
        )
        try:
            yield
        finally:
            process.ctrl_pop()


def ctrl(
    control_qubits: Quant, gate: Callable, state: int | list[int] | None = None
) -> Callable:
    """Add control qubits to a gate.

    Create a new callable that applies the given ``gate`` with the control qubits.

    :Usage:

    .. code-block:: python

        from ket import *

        p = Process()

        c = p.alloc(2)
        a, b = q.alloc(2)

        # CNOT c[0] a
        ctrl(c[0], X)(a)

        # Toffoli c[0] c[1] a
        ctrl(c, X)(a)

        # CSWAP c[0] a b
        ctrl(c[0], SWAP)(a, b)

    Args:
        control_qubits: The qubits to control the quantum operations.
        gate: The gate to apply with the control qubits.

    Returns:
        A new callable that applies the given gate with the control qubits.
    """

    @wraps(gate)
    def inner(*args, **kwargs):
        with control(control_qubits, state):
            return control_qubits, gate(*args, **kwargs)

    return inner


def C(gate: Callable) -> Callable:  # pylint: disable=invalid-name
    """Create a controlled gate

    Args:
        gate: Quantum gate.

    Returns:
        A gate with the first parameter as control.
    """

    @wraps(gate)
    def inner(control_qubits, *args, **kwargs):
        return ctrl(control_qubits, gate)(*args, **kwargs)

    return inner


def _search_process(ket_process, args, kwargs):
    def inner(ket_process, arg):
        if hasattr(arg, "_get_ket_process"):
            arg_process = arg._get_ket_process()  # pylint: disable=protected-access
            if ket_process is not None and ket_process is not arg_process:
                raise ValueError("parameter with different Ket processes")
            ket_process = arg_process
        return ket_process

    def search(ket_process, args):
        for arg in args:
            if isinstance(arg, Sequence) and not isinstance(arg, str):
                for subarg in arg:
                    ket_process = inner(ket_process, subarg)
            else:
                ket_process = inner(ket_process, arg)
        return ket_process

    if ket_process is None:
        ket_process = search(ket_process, args)
        ket_process = search(ket_process, kwargs.values())

    if ket_process is None:
        raise ValueError("Ket process not found in the parameters")

    return ket_process


def adj(gate: Callable[[Any], Any]) -> Callable[[Any], Any]:
    """Return the inverse of a gate.

    Create a new callable that applies the inverse of the given ``gate``.

    The resulting callable will iterate over the parameters to identify the
    :class:`~ket.base.Process` that the inverse operation will be applied. If non of the parameters
    are a :class:`~ket.base.Quant`, the keyword argument ``ket_process`` must be given.

    :Usage:

    .. code-block:: python

        from ket import *

        p = Process()

        a, b = q.alloc(2)

        bell = cat(kron(H, I), CNOT)

        adj(bell)(a, b)

    Args:
        gate: The gate to apply the inverse.

    Returns:
        A new callable that applies the inverse of the given gate.

    """

    @wraps(gate)
    def inner(*args, ket_process: Process | None = None, **kwargs) -> Any:
        ket_process = _search_process(ket_process, args, kwargs)

        with inverse(ket_process):
            return gate(*args, **kwargs)

    return inner


@contextmanager
def inverse(process: Process):
    """Inverse scope.

    Open a scope where all the quantum operations are executed backwards.

    :Usage:

    .. code-block:: python

       with inverse(process):
            # Classical and quantum operations
            ...

    Args:
        process: The process where the inverse operations will be applied.
    """
    process.adj_begin()
    try:
        yield
    finally:
        process.adj_end()


def cat(*gates) -> Callable[[Any], Any]:
    """Concatenate gates.

    Create a new callable concatenating all the quantum gates.

    Example:

        .. code-block:: python

            from ket import *

            z_gate = cat(H, X, H)

            p = process()
            q = p.alloc()

            z_gate(q)

    Args:
        *gates: Quantum gates to concatenate.

    Returns:
        A new callable that concatenates the given quantum gates.
    """

    def inner(*args):
        for gate in gates:
            gate(*args)

        if len(args) == 1:
            return args[0]
        return args

    return inner


def kron(*gates, n: int = 1) -> Callable[[Any], Any]:
    """Gates tensor product.

    Create a new callable that with the tensor product of the given gates.

    Example:

        .. code-block:: python

            from ket import *

            p = process()
            a, b = p.alloc(2)

            HX = kron(H, X)

            HX(a, b) # Apply an Hadamard on qubit `a` and a Pauli X on qubit `b`

    Args:
        *gates: Quantum gates to tensor product.

    Returns:
        A new callable that represents the tensor product of the given quantum gates.
    """

    gates = gates * n

    def inner(*args):
        return tuple(
            gate(*arg) if isinstance(arg, tuple) else gate(arg)
            for gate, arg in zip(gates, args)
        )

    return inner


@contextmanager
def around(gate: Callable, *args, ket_process: Process | None = None, **kwargs):
    r"""Applying and then reversing quantum gates.

    Apply the given quantum gate (:math:`U`) and then execute a code block (:math:`V`). After the
    code block is executed, the inverse of the gate (:math:`U^\dagger`) is applied, resulting in
    :math:`UVU^\dagger`.

    Example:

        .. code-block:: python

            from ket import *

            p = Process()
            a, b = p.alloc(2)

            bell = cat(kron(H, I), CNOT)

            with around(bell, a, b): # Apply bell(a, b)
                X(q)                 # Execute the code block
                                    # Apply adj(bell)(a, b)

    Args:
        gate: Quantum gate to apply.
        *arg: Qubits or classical parameters for the gate.
        ket_process: Explicitly specify the quantum process. If not provided, the process
                     is inferred from the qubits.
        **kwargs: Additional parameters for the quantum gate.

    """

    ket_process = _search_process(ket_process, args, kwargs)

    ket_process.around_begin()

    ket_process.ctrl_stack()
    # ket_process.approximated_decomposition_begin()
    gate(*args, **kwargs)
    # ket_process.approximated_decomposition_end()
    ket_process.ctrl_unstack()

    ket_process.around_mid()

    try:
        yield
    finally:

        ket_process.around_undo()

        ket_process.ctrl_stack()
        # ket_process.approximated_decomposition_begin()
        adj(gate)(*args, ket_process=ket_process, **kwargs)
        # ket_process.approximated_decomposition_end()
        ket_process.ctrl_unstack()

        ket_process.around_end()


def measure(qubits: Quant) -> Measurement:
    """Measure the given qubits and return a measurement object.

    Args:
        qubits: Qubits to be measured.

    Returns:
        Object representing the measurement results.
    """
    if not isinstance(qubits, Quant):
        qubits = reduce(add, qubits)

    return Measurement(qubits)


def dump(qubits: Quant) -> QuantumState:
    """Obtain the quantum state snapshot.

    Args:
        qubits: Qubits to be observed.

    Returns:
        Object representing the quantum state.
    """
    if not isinstance(qubits, Quant):
        qubits = reduce(add, qubits)

    return QuantumState(qubits)


def sample(qubits: Quant, shots: int = 2048) -> Samples:
    """Get the quantum state measurement samples.

    Args:
        qubits: Qubits to be measured.
        shots: Number of measurement shots.

    Returns:
        Object representing the measurement samples.
    """
    if not isinstance(qubits, Quant):
        qubits = reduce(add, qubits)

    return Samples(qubits, int(shots))


def exp_value(hamiltonian: Hamiltonian | Pauli) -> ExpValue:
    """Calculate the expected value for a quantum state.

    Args:
        hamiltonian: Hamiltonian or Pauli operator for calculating the expected value.

    Returns:
        Object representing the expected value.
    """
    return ExpValue(hamiltonian)


def using_aux(unsafe: bool = False, **names):
    """Add axillary qubits to a quantum gate.

    Example:
            .. code-block:: python

                @using_aux(a=lambda c: 0 if len(c) <= 2 else 1)
                def v_chain(c, t, a):
                    if len(c) <= 2:
                        ctrl(c, X)(t)
                    else:
                        with around(ctrl(c[:2], X), a):
                            v_chain(a + c[2:], t)

    Args:
        unsafe: Use dirty unsafe qubits. Defaults to False.
    """

    def inner(func):
        param = signature(func).parameters

        @wraps(func)
        def call(*args, ket_process: Process | None = None, **kwargs):
            ket_process = _search_process(ket_process, args, kwargs)

            if unsafe:
                q_args = [q for q in args if isinstance(q, Quant)] + [
                    q for q in kwargs.values() if isinstance(q, Quant)
                ]
                interacting_qubits = reduce(add, q_args)
                interacting_qubits = interacting_qubits.qubits
                interacting_qubits_size = len(interacting_qubits)
                interacting_qubits = (c_size_t * interacting_qubits_size)(
                    *interacting_qubits
                )
            else:
                interacting_qubits_size = 0
                interacting_qubits = None

            free_stack = []

            kwargs_ex = {**kwargs, **dict(zip(param, args))}

            for name, num_qubits in names.items():
                if name in kwargs_ex:
                    continue

                if callable(num_qubits):
                    try:
                        nq_args = {
                            p: kwargs_ex[p] for p in signature(num_qubits).parameters
                        }
                        num_qubits = num_qubits(**nq_args)
                    except KeyError as e:
                        raise ValueError(
                            "Parameters used to calculate the number of auxiliary "
                            "qubit must be passed as keyword arguments"
                        ) from e

                if num_qubits <= 0:
                    kwargs[name] = None
                    continue

                aux_index, aux_id = ket_process.allocate_aux(
                    num_qubits, interacting_qubits, interacting_qubits_size
                )

                qubits = Quant(
                    qubits=[
                        (i + 1) << 32
                        for i in range(aux_index.value, aux_index.value + num_qubits)
                    ],
                    process=ket_process,
                )

                free_stack.append(aux_id.value)

                kwargs[name] = qubits

            func(*args, **kwargs)

            for aux_id in reversed(free_stack):
                ket_process.free_aux(aux_id)

        return call

    return inner


def is_diagonal(gate: Callable) -> Callable:
    """Force to consider as a diagonal gate.

    Args:
        gate: Quantum gate
    """

    @wraps(gate)
    def inner(*args, ket_process: Process | None = None, **kwargs) -> Any:
        ket_process = _search_process(ket_process, args, kwargs)

        ket_process.is_diagonal_begin()
        try:
            return gate(*args, **kwargs)
        finally:
            ket_process.is_diagonal_end()

    return inner


def is_permutation(gate: Callable) -> Callable:
    """Force to consider as a permutation gate.

    Args:
        gate: Quantum gate
    """

    @wraps(gate)
    def inner(*args, ket_process: Process | None = None, **kwargs) -> Any:
        ket_process = _search_process(ket_process, args, kwargs)

        ket_process.is_permutation_begin()
        try:
            return gate(*args, **kwargs)
        finally:
            ket_process.is_permutation_end()

    return inner



"""Wrapper for Libket C API."""

from __future__ import annotations

# SPDX-FileCopyrightText: 2020 Evandro Chagas Ribeiro da Rosa <evandro@quantuloop.com>
# SPDX-FileCopyrightText: 2020 Rafael de Santiago <r.santiago@ufsc.br>
#
# SPDX-License-Identifier: Apache-2.0

from ctypes import (
    CFUNCTYPE,
    Structure,
    c_uint32,
    c_void_p,
    c_size_t,
    POINTER,
    c_bool,
    c_uint8,
    c_int32,
    c_uint64,
    c_double,
)
import json
from typing import Any, Literal
import weakref
from os import environ
from os.path import dirname
from abc import ABC, abstractmethod

from .wrapper import load_lib, os_lib_name


HADAMARD = 0
PAULI_X = 1
PAULI_Y = 2
PAULI_Z = 3
ROTATION_X = 10
ROTATION_Y = 20
ROTATION_Z = 30
PHASE_SHIFT = 31

SUCCESS = 0
CONTROL_TWICE = 1
DATA_NOT_AVAILABLE = 2
DEALLOCATED_QUBIT = 3
QUBIT_INDEX_OUT_OF_BOUNDS = 4
NUMBER_OF_QUBITS_EXCEEDED = 5
NO_ADJ = 6
NO_CTRL = 7
NON_GATE_INSTRUCTION_IN_ADJ = 8
TARGET_IN_CONTROL = 9
PROCESS_READY_TO_EXECUTE = 10
UNEXPECTED_RESULT_DATA = 11
DUMP_NOT_ALLOWED = 12
EXP_VALUE_NOT_ALLOWED = 13
SAMPLE_NOT_ALLOWED = 14
MEASURE_NOT_ALLOWED = 15
UNDEFINED_ERROR = 16


class BatchCExecution(Structure):  # pylint: disable=too-few-public-methods
    """C BatchCExecution Structure"""

    _fields_ = [
        (
            "submit_execution",
            CFUNCTYPE(
                None,
                POINTER(c_uint8),
                c_size_t,
                POINTER(c_double),
                c_size_t,
            ),
        ),
        ("get_results", CFUNCTYPE(None, POINTER(POINTER(c_uint8)), POINTER(c_size_t))),
        ("clear", CFUNCTYPE(None)),
    ]


class LiveCExecution(Structure):  # pylint: disable=too-few-public-methods
    """C LiveCExecution Structure"""

    _fields_ = [
        (
            "gate",
            CFUNCTYPE(
                None, POINTER(c_uint8), c_size_t, c_size_t, POINTER(c_size_t), c_size_t
            ),
        ),
        ("measure", CFUNCTYPE(c_uint64, POINTER(c_size_t), c_size_t)),
        ("exp_value", CFUNCTYPE(c_double, POINTER(c_uint8), c_size_t)),
        (
            "sample",
            CFUNCTYPE(
                None,
                POINTER(c_size_t),
                c_size_t,
                c_size_t,
                POINTER(POINTER(c_uint8)),
                POINTER(c_size_t),
            ),
        ),
        (
            "dump",
            CFUNCTYPE(
                None,
                POINTER(c_size_t),
                c_size_t,
                POINTER(POINTER(c_uint8)),
                POINTER(c_size_t),
            ),
        ),
        ("save", CFUNCTYPE(None, POINTER(POINTER(c_uint8)), POINTER(c_size_t))),
        ("load", CFUNCTYPE(None, POINTER(c_uint8), c_size_t)),
    ]


API_argtypes = {
    # 'ket_type_method': ([input_list], [output_list]),
    "ket_set_log_level": ([c_uint32], []),
    "ket_build_info": ([], [POINTER(c_uint8), c_size_t]),
    "ket_process_new": ([c_void_p], [c_void_p]),
    "ket_process_delete": ([c_void_p], []),
    "ket_process_allocate_qubit": ([c_void_p], [c_size_t]),
    "ket_process_apply_gate": (
        [
            c_void_p,  # process
            c_int32,  # gate
            c_double,  # angle
            c_bool,  # use_param
            c_size_t,  # param_index
            c_size_t,  # target
        ],
        [],
    ),
    "ket_process_set_parameter": ([c_void_p, c_double], [c_size_t]),
    "ket_process_get_gradient": ([c_void_p, c_size_t], [c_bool, c_double]),
    "ket_process_apply_global_phase": (
        [c_void_p, c_double],
        [],
    ),
    "ket_process_measure": ([c_void_p, POINTER(c_size_t), c_size_t], [c_size_t]),
    "ket_hamiltonian_new": ([], [c_void_p]),
    "ket_hamiltonian_add": (
        [c_void_p, POINTER(c_int32), c_size_t, POINTER(c_size_t), c_size_t, c_double],
        [],
    ),
    "ket_process_exp_value": (
        [c_void_p, c_void_p],
        [c_size_t],
    ),
    "ket_process_sample": (
        [c_void_p, POINTER(c_size_t), c_size_t, c_uint64],
        [c_size_t],
    ),
    "ket_process_dump": ([c_void_p, POINTER(c_size_t), c_size_t], [c_size_t]),
    "ket_process_ctrl_push": ([c_void_p, POINTER(c_size_t), c_size_t], []),
    "ket_process_ctrl_pop": ([c_void_p], []),
    "ket_process_ctrl_stack": ([c_void_p], []),
    "ket_process_ctrl_unstack": ([c_void_p], []),
    "ket_process_adj_begin": ([c_void_p], []),
    "ket_process_adj_end": ([c_void_p], []),
    "ket_process_execute": ([c_void_p], []),
    "ket_process_prepare_for_execution": ([c_void_p], []),
    "ket_process_instructions_json": (
        [c_void_p, POINTER(c_uint8), c_size_t],
        [c_size_t],
    ),
    "ket_process_isa_instructions_json": (
        [c_void_p, POINTER(c_uint8), c_size_t],
        [c_size_t],
    ),
    "ket_process_metadata_json": ([c_void_p, POINTER(c_uint8), c_size_t], [c_size_t]),
    "ket_process_get_measurement": ([c_void_p, c_size_t], [c_bool, c_uint64]),
    "ket_process_get_exp_value": ([c_void_p, c_size_t], [c_bool, c_double]),
    "ket_process_get_sample": (
        [c_void_p, c_size_t],
        [c_bool, POINTER(c_uint64), POINTER(c_uint64), c_size_t],
    ),
    "ket_process_get_dump_size": ([c_void_p, c_size_t], [c_bool, c_size_t]),
    "ket_process_get_dump": (
        [c_void_p, c_size_t, c_size_t],
        [POINTER(c_uint64), c_size_t, c_double, c_double],
    ),
    "ket_make_configuration": (
        [
            POINTER(c_uint8),  # json
            c_size_t,  # json_size
            POINTER(BatchCExecution),  # batch_execution
            POINTER(LiveCExecution),  # live_execution
        ],
        [c_void_p],
    ),
    "ket_process_save_sim_state": ([c_void_p, POINTER(c_uint8), c_size_t], [c_size_t]),
    "ket_process_load_sim_state": ([c_void_p, POINTER(c_uint8), c_size_t], []),
    "ket_process_is_diagonal_begin": ([c_void_p], []),
    "ket_process_is_permutation_begin": ([c_void_p], []),
    "ket_process_is_diagonal_end": ([c_void_p], []),
    "ket_process_is_permutation_end": ([c_void_p], []),
    "ket_process_around_begin": ([c_void_p], []),
    "ket_process_around_mid": ([c_void_p], []),
    "ket_process_around_undo": ([c_void_p], []),
    "ket_process_around_end": ([c_void_p], []),
    "ket_process_allocate_aux": (
        [c_void_p, c_size_t, POINTER(c_size_t), c_size_t],
        [c_size_t, c_size_t],
    ),
    "ket_process_free_aux": (
        [c_void_p, c_size_t],
        [],
    ),
    "ket_process_approximated_decomposition_begin": ([c_void_p], []),
    "ket_process_approximated_decomposition_end": ([c_void_p], []),
}


def libket_path():
    """Get Libket shared library path"""

    if "LIBKET_PATH" in environ:
        path = environ["LIBKET_PATH"]
    else:
        path = f'{dirname(__file__)}/libs/{os_lib_name("ket")}'

    return path


API = load_lib("Libket", libket_path(), API_argtypes, "ket_error_message")


def set_log(level: int):
    """Set Libket log level"""

    API["ket_set_log_level"](level)


class Process:
    """Libket process wrapper from C API"""

    def __init__(self, configuration):
        self._as_parameter_ = API["ket_process_new"](configuration)
        self._finalizer = weakref.finalize(
            self, API["ket_process_delete"], self._as_parameter_
        )

    def __getattr__(self, name: str):
        return lambda *args: API["ket_process_" + name](self, *args)

    def __repr__(self) -> str:
        return "<Libket 'process'>"


class BatchExecution(ABC):
    """Base class for constructing batch target executions."""

    def __init__(self):
        @CFUNCTYPE(
            None,
            POINTER(c_uint8),
            c_size_t,
            POINTER(c_double),
            c_size_t,
        )
        def submit_execution(  # pylint: disable=too-many-arguments,too-many-positional-arguments
            circuit,
            circuit_size,
            parameters,
            parameters_size,
        ):
            circuit = json.loads(bytearray(circuit[:circuit_size]))

            parameters = parameters[:parameters_size]
            self.submit_execution(circuit, parameters)

        self._result_json = None
        self._result_len = None

        @CFUNCTYPE(None, POINTER(POINTER(c_uint8)), POINTER(c_size_t))
        def get_result(result_ptr, size):
            result = self.get_result()
            self._result_json = json.dumps(result).encode("utf-8")
            self._result_len = len(self._result_json)
            self._result_json = (c_uint8 * self._result_len)(*self._result_json)
            result_ptr[0] = self._result_json
            size[0] = self._result_len

        @CFUNCTYPE(None)
        def clear():
            self.clear()

        self.c_struct = BatchCExecution(
            submit_execution,
            get_result,
            clear,
        )

    @abstractmethod
    def submit_execution(
        self,
        circuit: dict,
        parameters: list[float],
    ):
        """Get the quantum circuit to execute.

        .. warning::
            This method is called by Libket and should not be called directly.
        """

    @abstractmethod
    def get_result(self) -> dict:
        """Get the result of the quantum circuit execution.

        .. warning::
            This method is called by Libket and should not be called directly.
        """

    @abstractmethod
    def clear(self):
        """Clear the data to start a new execution.

        .. warning::
            This method is called by Libket and should not be called directly.
        """

    @abstractmethod
    def connect(self):
        """Call :meth:`ket.clib.libket.BatchExecution.configure` with the
        appropriated arguments to generate the quantum execution target.

        .. warning::
            This method is called automatically by :class:`~ket.base.Process`.
            It is not necessary to call it manually.

        """

    def configure(self, **kwargs):
        """Configure the batch execution."""
        return make_configuration(execution=self, **kwargs)

    @staticmethod
    def get_gate_and_angle(gate):
        """Get the gate and possible angle from the gate object."""
        if isinstance(gate, str):
            return gate, {}
        return list(gate.items())[0]

    def process_instructions(self, instructions: dict):
        """Parse the instructions to execute."""
        for instruction in instructions:
            match instruction:
                case {"Gate": {"control": control, "gate": gate, "target": target}}:
                    gate, param = self.get_gate_and_angle(gate)
                    match gate:
                        case "Hadamard":
                            self.hadamard(target, control, **param)
                        case "PauliX":
                            self.pauli_x(target, control, **param)
                        case "PauliY":
                            self.pauli_y(target, control, **param)
                        case "PauliZ":
                            self.pauli_z(target, control, **param)
                        case "RotationX":
                            self.rotation_x(target, control, **param)
                        case "RotationY":
                            self.rotation_y(target, control, **param)
                        case "RotationZ":
                            self.rotation_z(target, control, **param)
                        case "Phase":
                            self.phase(target, control, **param)
                case {"ExpValue": {"index": index, "hamiltonian": hamiltonian}}:
                    self.exp_value(index, hamiltonian)
                case {"Measure": {"index": index, "qubits": qubits}}:
                    self.measure(index, qubits)
                case {"Sample": {"index": index, "qubits": qubits, "shots": shots}}:
                    self.sample(index, qubits, shots)
                case {"Dump": {"index": index, "qubits": qubits}}:
                    self.dump(index, qubits)
                case "Identity":
                    continue
                case _:
                    raise ValueError(f"Invalid instruction: {instruction}")

    def pauli_x(self, target, control):
        """Apply a Pauli-X gate to the target qubit.

        .. warning::
            This method is called by Libket and  should not be called directly.

        """
        raise NotImplementedError("Pauli-X gate not implemented")

    def pauli_y(self, target, control):
        """Apply a Pauli-Y gate to the target qubit.

        .. warning::
            This method is called by Libket and should not be called directly.
        """
        raise NotImplementedError("Pauli-Y gate not implemented")

    def pauli_z(self, target, control):
        """Apply a Pauli-Z gate to the target qubit.

        .. warning::
            This method is called by Libket and should not be called directly.
        """
        raise NotImplementedError("Pauli-Z gate not implemented")

    def hadamard(self, target, control):
        """Apply a Hadamard gate to the target qubit.

        .. warning::
            This method is called by Libket and should not be called directly.
        """
        raise NotImplementedError("Hadamard gate not implemented")

    def rotation_x(self, target, control, **kwargs):
        """Apply a X-Rotation gate to the target qubit.

        .. warning::
            This method is called by Libket and should not be called directly.
        """
        raise NotImplementedError("X-Rotation gate not implemented")

    def rotation_y(self, target, control, **kwargs):
        """Apply a Y-Rotation gate to the target qubit.

        .. warning::
            This method is called by Libket and should not be called directly.
        """
        raise NotImplementedError("Y-Rotation gate not implemented")

    def rotation_z(self, target, control, **kwargs):
        """Apply a Z-Rotation gate to the target qubit.

        .. warning::
            This method is called by Libket and should not be called directly.
        """
        raise NotImplementedError("Z-Rotation gate not implemented")

    def phase(self, target, control, **kwargs):
        """Apply a Phase gate to the target qubit.

        .. warning::
            This method is called by Libket and should not be called directly.
        """
        raise NotImplementedError("Phase gate not implemented")

    def exp_value(self, index, hamiltonian):
        """Compute the expectation value.

        .. warning::
            This method is called by Libket and should not be called directly.
        """
        raise NotImplementedError("Expectation value not implemented")

    def measure(self, index, qubits) -> int:
        """Measure the qubits.

        .. warning::
            This method is called by Libket and should not be called directly.
        """
        raise NotImplementedError("Measurement not implemented")

    def sample(self, index, qubits, shots):
        """Sample the qubits.

        .. warning::
            This method is called by Libket and should not be called directly.
        """
        raise NotImplementedError("Sampling not implemented")

    def dump(self, index, qubits):
        """Dump the qubits.

        .. warning::
            This method is called by Libket and should not be called directly.
        """
        raise NotImplementedError("Dump not implemented")


class LiveExecution(ABC):
    """Base class for constructing live target executions."""

    def __init__(self):
        @CFUNCTYPE(
            None, POINTER(c_uint8), c_size_t, c_size_t, POINTER(c_size_t), c_size_t
        )
        def gate(gate_json, gate_json_size, target, control, control_size):
            gate = json.loads(bytearray(gate_json[:gate_json_size]))
            control = control[:control_size]
            if isinstance(gate, dict):
                gate, value = list(gate.items())[0]
                value = value["Value"]
            match gate:
                case "Hadamard":
                    self.hadamard(target, control)
                case "PauliX":
                    self.pauli_x(target, control)
                case "PauliY":
                    self.pauli_y(target, control)
                case "PauliZ":
                    self.pauli_z(target, control)
                case "RotationX":
                    self.rotation_x(target, control, value)
                case "RotationY":
                    self.rotation_y(target, control, value)
                case "RotationZ":
                    self.rotation_z(target, control, value)
                case "Phase":
                    self.phase(target, control, value)

        @CFUNCTYPE(c_uint64, POINTER(c_size_t), c_size_t)
        def measure(qubits, qubits_size):
            qubits = qubits[:qubits_size]
            return self.measure(qubits)

        @CFUNCTYPE(c_double, POINTER(c_uint8), c_size_t)
        def exp_value(h_json, h_size):
            h = json.loads(bytearray(h_json[:h_size]))
            return self.exp_value(h)

        self._sample_json = None
        self._sample_len = None

        @CFUNCTYPE(
            None,
            POINTER(c_size_t),
            c_size_t,
            c_size_t,
            POINTER(POINTER(c_uint8)),
            POINTER(c_size_t),
        )
        def sample(qubits, qubits_size, shots, result_ptr, size):
            qubits = qubits[:qubits_size]
            result = self.sample(qubits, shots)
            self._sample_json = json.dumps(result).encode("utf-8")
            self._sample_len = len(self._sample_json)
            self._sample_json = (c_uint8 * self._sample_len)(*self._sample_json)
            result_ptr[0] = self._sample_json
            size[0] = self._sample_len

        self._dump_json = None
        self._dump_len = None

        @CFUNCTYPE(
            None,
            POINTER(c_size_t),
            c_size_t,
            POINTER(POINTER(c_uint8)),
            POINTER(c_size_t),
        )
        def dump(qubits, qubits_size, result_ptr, size):
            qubits = qubits[:qubits_size]
            result = self.dump(qubits)
            self._dump_json = json.dumps(result).encode("utf-8")
            self._dump_len = len(self._dump_json)
            self._dump_json = (c_uint8 * self._dump_len)(*self._dump_json)
            result_ptr[0] = self._dump_json
            size[0] = self._dump_len

        self._save_data = None
        self._save_len = None

        @CFUNCTYPE(None, POINTER(POINTER(c_uint8)), POINTER(c_size_t))
        def save(save_data, save_len):
            self._save_data = self.save()
            self._sample_len = len(self._save_data)
            self._save_data = (c_uint8 * self._save_len)(*self._save_data)
            save_data[0] = self._save_data
            save_len[0] = self._save_len

        @CFUNCTYPE(None, POINTER(c_uint8), c_size_t)
        def load(data, size):
            data = bytearray(data[:size])
            self.load(data)

        self.c_struct = LiveCExecution(
            gate,
            measure,
            exp_value,
            sample,
            dump,
            save,
            load,
        )

    def pauli_x(self, target, control):
        """Apply a Pauli-X gate to the target qubit."""
        raise NotImplementedError("Pauli-X gate not implemented")

    def pauli_y(self, target, control):
        """Apply a Pauli-Y gate to the target qubit."""
        raise NotImplementedError("Pauli-Y gate not implemented")

    def pauli_z(self, target, control):
        """Apply a Pauli-Z gate to the target qubit."""
        raise NotImplementedError("Pauli-Z gate not implemented")

    def hadamard(self, target, control):
        """Apply a Hadamard gate to the target qubit."""
        raise NotImplementedError("Hadamard gate not implemented")

    def rotation_x(self, target, control, angle):
        """Apply a X-Rotation gate to the target qubit."""
        raise NotImplementedError("X-Rotation gate not implemented")

    def rotation_y(self, target, control, angle):
        """Apply a Y-Rotation gate to the target qubit."""
        raise NotImplementedError("Y-Rotation gate not implemented")

    def rotation_z(self, target, control, angle):
        """Apply a Z-Rotation gate to the target qubit."""
        raise NotImplementedError("Z-Rotation gate not implemented")

    def phase(self, target, control, angle):
        """Apply a Phase gate to the target qubit."""
        raise NotImplementedError("Phase gate not implemented")

    def measure(self, qubits: list[int]) -> int:
        """Measure the qubits and return the result as an integer."""
        raise NotImplementedError("`measure` method not implemented")

    def exp_value(self, hamiltonian: dict):
        """Compute the expectation value of the given Hamiltonian."""
        raise NotImplementedError("`exp_value` method not implemented")

    def sample(self, qubits: list[int], shots: int):
        """Sample the state of the qubits."""
        raise NotImplementedError("`sample` method not implemented")

    def dump(self, qubits: list[int]) -> dict:
        """Dump the state of the qubits."""
        raise NotImplementedError("`dump` method not implemented")

    def save(self) -> bytearray:
        """Save the current state of the execution."""
        raise NotImplementedError("`save` method not implemented")

    def load(self, data: bytearray):
        """Load the state of the execution from the given data."""
        raise NotImplementedError("`load` method not implemented")

    @abstractmethod
    def connect(self):
        """Call configure with the appropriated arguments to generate the object."""

    def configure(self, **kwargs):
        """Configure the batch execution."""
        return make_configuration(execution=self, **kwargs)


_BASE_QPU = {
    "coupling_graph": None,
    "u2_gates": "All",  # All, ZYZ, RzSx
    "u4_gate": "CX",  # CX, CZ
}

# Unsupported, Basic, Advanced
_MANAGED_BY_TARGET = {
    "measure": "Unsupported",
    "sample": "Unsupported",
    "exp_value": "Unsupported",
    "dump": "Unsupported",
}

_CLASSICAL_SHADOWS = {"bias": (1, 1, 1), "samples": 1_000, "shots": 2048}


def _dfs(node, visited, adj):
    visited.add(node)
    for neighbor in adj.get(node, []):
        if neighbor not in visited:
            _dfs(neighbor, visited, adj)


def _find_unreachable_qubits(num_qubits, edges):
    adj = {}
    for u, v in edges:
        adj.setdefault(u, []).append(v)
        adj.setdefault(v, []).append(u)

    all_nodes = set(range(num_qubits))
    visited = set()
    _dfs(0, visited, adj)

    return list(all_nodes - visited)


def make_configuration(  # pylint: disable=too-many-arguments,too-many-positional-arguments
    num_qubits: int,
    execution,
    execution_managed_by_target: dict[str, str] | None = None,
    direct_sample_exp_value: int | None = None,
    classical_shadows_exp_value: dict[str, Any] | None = None,
    qpu: dict[str, str] | None = None,
    gradient: Literal["ParameterShift", "NativeSupport"] | None = None,
) -> Process:
    """Make a Libket configuration"""

    if qpu is not None:
        qpu = {**_BASE_QPU, **qpu}
        if qpu["coupling_graph"]:
            unreachable_qubits = _find_unreachable_qubits(
                num_qubits, qpu["coupling_graph"]
            )
            if len(unreachable_qubits) > 0:
                raise ValueError(
                    f"Unreachable qubit {unreachable_qubits} in the coupling graph."
                )

    # Ensure only one of the three is defined and at least one is provided
    defined_options = [
        execution_managed_by_target is not None,
        direct_sample_exp_value is not None,
        classical_shadows_exp_value is not None,
    ]
    if sum(defined_options) != 1:
        raise ValueError(
            "Exactly one of 'execution_managed_by_target', 'direct_sample_exp_value', "
            "or 'classical_shadows_exp_value' must be defined."
        )

    if execution_managed_by_target is not None:
        execution_protocol = {
            "ManagedByTarget": {**_MANAGED_BY_TARGET, **execution_managed_by_target}
        }
    elif direct_sample_exp_value is not None:
        execution_protocol = {"SampleBased": {"DirectSample": direct_sample_exp_value}}
    else:
        execution_protocol = {
            "SampleBased": {
                "ClassicalShadows": {
                    **_CLASSICAL_SHADOWS,
                    **classical_shadows_exp_value,
                }
            }
        }

    execution_target = {
        "num_qubits": num_qubits,
        "qpu": qpu,
        "execution_protocol": execution_protocol,
        "gradient": gradient,
    }

    execution_target_json = json.dumps(execution_target).encode("utf-8")
    execution_target_len = len(execution_target_json)
    execution_target_json = (c_uint8 * execution_target_len)(*execution_target_json)

    if isinstance(execution, BatchExecution):
        batch_execution = execution.c_struct
        live_execution = None
    elif isinstance(execution, LiveExecution):
        batch_execution = None
        live_execution = execution.c_struct
    else:
        batch_execution = None
        live_execution = None

    return API["ket_make_configuration"](
        execution_target_json,
        execution_target_len,
        batch_execution,
        live_execution,
    )
